from ape import plugins
from ape.api import NetworkAPI, create_network_type

from .converters import WeiConversions
from .ecosystem import NETWORKS, Ethereum


@plugins.register(plugins.ConversionPlugin)
def converters():
    yield int, WeiConversions


@plugins.register(plugins.EcosystemPlugin)
def ecosystems():
    yield Ethereum


@plugins.register(plugins.NetworkPlugin)
def networks():
    for network_name, network_params in NETWORKS.items():
        yield "ethereum", network_name, create_network_type(*network_params)

    # NOTE: This works for `geth --dev` as it gets chain_id from itself
    yield "ethereum", "development", NetworkAPI



from pathlib import Path
from typing import Dict, List, Set

from dataclassy import dataclass

from ape.api.compiler import CompilerAPI
from ape.logging import logger
from ape.plugins import PluginManager
from ape.types import ContractType
from ape.utils import cached_property

from ..exceptions import CompilerError
from .config import ConfigManager


@dataclass
class CompilerManager:
    config: ConfigManager
    plugin_manager: PluginManager

    @cached_property
    def registered_compilers(self) -> Dict[str, CompilerAPI]:
        registered_compilers = {}

        for plugin_name, (extensions, compiler_class) in self.plugin_manager.register_compiler:
            # TODO: Add config via ``self.config.get_config(plugin_name)``
            compiler = compiler_class()

            for extension in extensions:

                if extension in registered_compilers:
                    raise CompilerError(f"Compiler for '{extension}' is already registered")

                registered_compilers[extension] = compiler

        return registered_compilers

    def compile(self, contract_filepaths: List[Path]) -> Dict[str, ContractType]:
        extensions = self._get_contract_extensions(contract_filepaths)

        contract_types = {}
        for extension in extensions:
            paths_to_compile = [path for path in contract_filepaths if path.suffix == extension]
            for path in paths_to_compile:
                logger.info(f"Compiling '{self._get_contract_path(path)}'")

            for contract_type in self.registered_compilers[extension].compile(paths_to_compile):

                if contract_type.contractName in contract_types:
                    raise CompilerError("ContractType collision across compiler plugins")

                contract_types[contract_type.contractName] = contract_type

        return contract_types

    def _get_contract_extensions(self, contract_filepaths: List[Path]) -> Set[str]:
        extensions = set(path.suffix for path in contract_filepaths)
        unhandled_extensions = extensions - set(self.registered_compilers)
        if len(unhandled_extensions) > 0:
            unhandled_extensions_str = ", ".join(unhandled_extensions)
            raise CompilerError(f"No compiler found for extensions [{unhandled_extensions_str}]")

        return extensions

    def _get_contract_path(self, path: Path):
        try:
            return path.relative_to(self.config.PROJECT_FOLDER)
        except ValueError:
            return path



import os
import subprocess
import sys
from pathlib import Path
from typing import Set

import click
from github import Github

from ape import config
from ape.cli import ape_cli_context, skip_confirmation_option
from ape.logging import logger
from ape.plugins import clean_plugin_name, plugin_manager
from ape.utils import get_package_version

# Plugins included with ape core
FIRST_CLASS_PLUGINS: Set[str] = {
    d.name for d in Path(__file__).parent.parent.iterdir() if d.is_dir and d.name.startswith("ape_")
}

# Plugins maintained OSS by ApeWorX (and trusted)
SECOND_CLASS_PLUGINS: Set[str] = set()

# TODO: Should the github client be in core?
# TODO: Handle failures with connecting to github (potentially cached on disk?)
if "GITHUB_ACCESS_TOKEN" in os.environ:
    author = Github(os.environ["GITHUB_ACCESS_TOKEN"]).get_organization("ApeWorX")

    SECOND_CLASS_PLUGINS = {
        repo.name.replace("-", "_") for repo in author.get_repos() if repo.name.startswith("ape-")
    }

else:
    logger.warning("$GITHUB_ACCESS_TOKEN not set, skipping 2nd class plugins")


def is_plugin_installed(plugin: str) -> bool:
    try:
        __import__(plugin)
        return True
    except ImportError:
        return False


@click.group(short_help="Manage ape plugins")
def cli():
    """
    Command-line helper for managing installed plugins.
    """


@cli.command(name="list", short_help="List installed plugins")
@click.option(
    "-a",
    "--all",
    "display_all",
    default=False,
    is_flag=True,
    help="Display all plugins (including Core)",
)
@ape_cli_context()
def _list(cli_ctx, display_all):
    plugins = []
    for name, plugin in plugin_manager.list_name_plugin():
        version_str = ""
        version = get_package_version(name)

        if name in FIRST_CLASS_PLUGINS:
            if not display_all:
                continue  # NOTE: Skip 1st class plugins unless specified
            version_str = " (core)"

        elif version:
            version_str = f" ({version})"

        plugins.append(f"{name}{version_str}")

    if plugins:
        click.echo("Installed plugins:")
        click.echo("  " + "\n  ".join(plugins))

    else:
        cli_ctx.logger.info("No plugins installed")


@cli.command(short_help="Install an ape plugin")
@click.argument("plugin")
@click.option("-v", "--version", help="Specify version (Default is latest)")
@skip_confirmation_option(help="Don't ask for confirmation to add the plugin")
@ape_cli_context()
def add(cli_ctx, plugin, version, skip_confirmation):
    if plugin.startswith("ape"):
        cli_ctx.abort(f"Namespace 'ape' in '{plugin}' is not required")

    # NOTE: Add namespace prefix (prevents arbitrary installs)
    plugin = f"ape_{clean_plugin_name(plugin)}"

    if version:
        plugin = f"{plugin}=={version}"

    if plugin in FIRST_CLASS_PLUGINS:
        cli_ctx.abort(f"Cannot add 1st class plugin '{plugin}'")

    elif is_plugin_installed(plugin):
        cli_ctx.abort(f"Plugin '{plugin}' already installed")

    elif (
        plugin in SECOND_CLASS_PLUGINS
        or skip_confirmation
        or click.confirm(f"Install unknown 3rd party plugin '{plugin}'?")
    ):
        cli_ctx.logger.info(f"Installing {plugin}...")
        # NOTE: Be *extremely careful* with this command, as it modifies the user's
        #       installed packages, to potentially catastrophic results
        # NOTE: This is not abstracted into another function *on purpose*
        subprocess.call([sys.executable, "-m", "pip", "install", "--quiet", plugin])


@cli.command(short_help="Install all plugins in the local config file")
@skip_confirmation_option("Don't ask for confirmation to install the plugins")
@ape_cli_context()
def install(cli_ctx, skip_confirmation):
    for plugin, version in config.get_config("plugins").items():
        if not plugin.startswith("ape-"):
            cli_ctx.abort(f"Namespace 'ape' required in config item '{plugin}'")

        if not is_plugin_installed(plugin.replace("-", "_")) and (
            plugin.replace("-", "_") in SECOND_CLASS_PLUGINS
            or skip_confirmation
            or click.confirm(f"Install unknown 3rd party plugin '{plugin}'?")
        ):
            cli_ctx.logger.info(f"Installing {plugin}...")
            # NOTE: Be *extremely careful* with this command, as it modifies the user's
            #       installed packages, to potentially catastrophic results
            # NOTE: This is not abstracted into another function *on purpose*
            subprocess.call(
                [sys.executable, "-m", "pip", "install", "--quiet", f"{plugin}=={version}"]
            )


@cli.command(short_help="Uninstall an ape plugin")
@click.argument("plugin")
@skip_confirmation_option("Don't ask for confirmation to remove the plugin")
@ape_cli_context()
def remove(cli_ctx, plugin, skip_confirmation):
    if plugin.startswith("ape"):
        cli_ctx.abort(f"Namespace 'ape' in '{plugin}' is not required")

    # NOTE: Add namespace prefix (match behavior of ``install``)
    plugin = f"ape_{clean_plugin_name(plugin)}"

    if not is_plugin_installed(plugin):
        cli_ctx.abort(f"Plugin '{plugin}' is not installed")

    elif plugin in FIRST_CLASS_PLUGINS:
        cli_ctx.abort(f"Cannot remove 1st class plugin '{plugin}'")

    elif skip_confirmation or click.confirm(
        f"Remove plugin '{plugin} ({get_package_version(plugin)})'"
    ):
        # NOTE: Be *extremely careful* with this command, as it modifies the user's
        #       installed packages, to potentially catastrophic results
        # NOTE: This is not abstracted into another function *on purpose*
        subprocess.call([sys.executable, "-m", "pip", "uninstall", "--quiet", "-y", plugin])



import pytest  # type: ignore
from ape_ethereum.converters import ETHER_UNITS
from eth_typing import ChecksumAddress
from hypothesis import given
from hypothesis import strategies as st

from ape import convert
from ape.exceptions import ConversionError


@pytest.mark.fuzzing
@given(
    value=st.decimals(
        min_value=-(2 ** 255),
        max_value=2 ** 256 - 1,
        allow_infinity=False,
        allow_nan=False,
    ),
    unit=st.sampled_from(list(ETHER_UNITS.keys())),
)
def test_ether_conversions(value, unit):
    actual = convert(value=f"{value} {unit}", type=int)
    expected = int(value * ETHER_UNITS[unit])
    assert actual == expected


def test_bad_type():
    with pytest.raises(ConversionError) as err:
        convert(value="something", type=float)

    expected = "Type '<class 'float'>' must be one of [ChecksumAddress, bytes, int]"
    assert str(err.value) == expected


def test_no_registered_converter():
    with pytest.raises(ConversionError) as err:
        convert(value="something", type=ChecksumAddress)

    assert str(err.value) == "No conversion registered to handle 'something'"



from typing import Any, Dict, List, Type

from dataclassy import dataclass
from eth_utils import is_checksum_address, is_hex, is_hex_address, to_checksum_address
from hexbytes import HexBytes

from ape.api import AddressAPI, ConverterAPI
from ape.exceptions import ConversionError
from ape.logging import logger
from ape.plugins import PluginManager
from ape.types import AddressType
from ape.utils import cached_property

from .config import ConfigManager
from .networks import NetworkManager


# NOTE: This utility converter ensures that all bytes args can accept hex too
class HexConverter(ConverterAPI):
    def is_convertible(self, value: str) -> bool:
        return is_hex(value)

    def convert(self, value: str) -> bytes:
        return HexBytes(value)


hex_converter = HexConverter(None, None)  # type: ignore


class AddressAPIConverter(ConverterAPI):
    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, AddressAPI)

    def convert(self, value: AddressAPI) -> AddressType:
        return value.address


address_api_converter = AddressAPIConverter(None, None)  # type: ignore


class HexAddressConverter(ConverterAPI):
    def is_convertible(self, value: str) -> bool:
        return isinstance(value, str) and is_hex_address(value) and not is_checksum_address(value)

    def convert(self, value: str) -> AddressType:
        logger.warning(f"'{value}' is not in checksummed form")
        return to_checksum_address(value)


hex_address_converter = HexAddressConverter(None, None)  # type: ignore


@dataclass
class ConversionManager:
    config: ConfigManager
    plugin_manager: PluginManager
    networks: NetworkManager

    @cached_property
    def _converters(self) -> Dict[Type, List[ConverterAPI]]:
        converters: Dict[Type, List[ConverterAPI]] = {
            AddressType: [address_api_converter, hex_address_converter],
            bytes: [hex_converter],
            int: [],
        }

        for plugin_name, (conversion_type, converter_class) in self.plugin_manager.converters:
            converter = converter_class(self.config.get_config(plugin_name), self.networks)

            if conversion_type not in converters:
                options = ", ".join([t.__name__ for t in converters])
                raise ConversionError(f"Type '{conversion_type}' must be one of [{options}]")

            converters[conversion_type].append(converter)

        return converters

    def is_type(self, value: Any, type: Type) -> bool:
        if type is AddressType:
            return is_checksum_address(value)

        else:
            return isinstance(value, type)

    def convert(self, value: Any, type: Type) -> Any:
        if type not in self._converters:
            options = ", ".join([t.__name__ for t in self._converters])
            raise ConversionError(f"Type '{type}' must be one of [{options}]")

        if self.is_type(value, type):
            return value

        for converter in self._converters[type]:
            if converter.is_convertible(value):
                return converter.convert(value)

        raise ConversionError(f"No conversion registered to handle '{value}'")



from typing import Iterator

from web3 import HTTPProvider, Web3
from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.middleware import geth_poa_middleware

from ape.api import ProviderAPI, ReceiptAPI, TransactionAPI
from ape.api.config import ConfigItem
from ape.exceptions import ProviderError

DEFAULT_SETTINGS = {"uri": "http://localhost:8545"}


class EthereumNetworkConfig(ConfigItem):
    # Make sure you are running the right networks when you try for these
    mainnet: dict = DEFAULT_SETTINGS.copy()
    ropsten: dict = DEFAULT_SETTINGS.copy()
    rinkeby: dict = DEFAULT_SETTINGS.copy()
    kovan: dict = DEFAULT_SETTINGS.copy()
    goerli: dict = DEFAULT_SETTINGS.copy()
    # Make sure to run via `geth --dev` (or similar)
    development: dict = DEFAULT_SETTINGS.copy()


class NetworkConfig(ConfigItem):
    ethereum: EthereumNetworkConfig = EthereumNetworkConfig()


class EthereumProvider(ProviderAPI):
    _web3: Web3 = None  # type: ignore

    @property
    def uri(self) -> str:
        return getattr(self.config, self.network.ecosystem.name)[self.network.name]["uri"]

    def connect(self):
        self._web3 = Web3(HTTPProvider(self.uri))
        self._web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)

        if self.network.name not in ("mainnet", "ropsten"):
            self._web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        if self.network.name != "development" and self.network.chain_id != self.chain_id:
            raise ProviderError(
                "HTTP Connection does not match expected chain ID. "
                f"Are you connected to '{self.network.name}'?"
            )

    def disconnect(self):
        self._web3 = None

    def update_settings(self, new_settings: dict):
        self.disconnect()
        self.provider_settings.update(new_settings)
        self.connect()

    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        return self._web3.eth.estimate_gas(txn.as_dict())  # type: ignore

    @property
    def chain_id(self) -> int:
        return self._web3.eth.chain_id

    @property
    def gas_price(self):
        return self._web3.eth.generate_gas_price()

    def get_nonce(self, address: str) -> int:
        return self._web3.eth.getTransactionCount(address)  # type: ignore

    def get_balance(self, address: str) -> int:
        return self._web3.eth.getBalance(address)  # type: ignore

    def get_code(self, address: str) -> bytes:
        return self._web3.eth.getCode(address)  # type: ignore

    def send_call(self, txn: TransactionAPI) -> bytes:
        data = txn.encode()
        return self._web3.eth.call(data)

    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        # TODO: Work on API that let's you work with ReceiptAPI and re-send transactions
        receipt = self._web3.eth.wait_for_transaction_receipt(txn_hash)  # type: ignore
        txn = self._web3.eth.get_transaction(txn_hash)  # type: ignore
        return self.network.ecosystem.receipt_class.decode({**txn, **receipt})

    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        txn_hash = self._web3.eth.send_raw_transaction(txn.encode())
        return self.get_transaction(txn_hash.hex())

    def get_events(self, **filter_params) -> Iterator[dict]:
        return iter(self._web3.eth.get_logs(filter_params))  # type: ignore



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from typing import List

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "ape"
copyright = "2021, ApeWorX LTD"
author = "ApeWorX LTD"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_multiversion",
    "sphinx_rtd_theme",
    "autoapi.extension",
]
autoapi_type = "python"
autoapi_dirs = ["../src"]
autoapi_template_dir = "_autoapi_templates"
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns: List[str] = ["_autoapi_templates/python/*.rst"]


def autoapi_skip_members(app, what, name, obj, skip, options):
    if "harambe" in name:
        return True
    return False


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", autoapi_skip_members)


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def fixpath(p):
    """
    Used to rewrite relative CSS paths to absolute paths in templates so we can
    get away with having only one copy of the CSS and web fonts.
    """
    new = f"/{project}/latest/_static/" + p.split("_static")[1]
    return new


html_context = {
    "fixpath": fixpath,
}



from enum import Enum
from typing import Any, Dict, Union

from .base import dataclass


class ConfigEnum(str, Enum):
    pass


@dataclass(slots=True, kwargs=True)
class ConfigItem:
    """
    Each plugin must inherit from this Config base class
    """

    def serialize(self) -> Dict:
        data: Dict[str, Union[str, int, Dict]] = dict()
        for name in self.__slots__:
            value = getattr(self, name)
            if isinstance(value, ConfigItem):
                data[name] = value.serialize()
            elif isinstance(value, ConfigEnum):
                data[name] = value.name
            elif isinstance(value, (int, str, dict)):
                data[name] = value
            else:
                raise TypeError(
                    f"Received unknown type '{type(data[name])}' when serializing a config item"
                )
        return data

    def validate_config(self):
        pass

    def __getitem__(self, attrname: str) -> Any:
        if attrname in self.__slots__:
            return getattr(self, attrname)

        raise KeyError(f"{attrname!r}")


class ConfigDict(ConfigItem):
    def __post_init__(self):
        raise ValueError("Do not use this class directly!")



from ape import plugins

from .providers import EthereumNetworkConfig, EthereumProvider, NetworkConfig


@plugins.register(plugins.Config)
def config_class():
    return NetworkConfig


@plugins.register(plugins.ProviderPlugin)
def providers():
    for network_name in EthereumNetworkConfig().serialize():
        yield "ethereum", network_name, EthereumProvider



from pathlib import Path
from tempfile import mkdtemp

import pytest  # type: ignore

import ape

# NOTE: Ensure that we don't use local paths for these
ape.config.DATA_FOLDER = Path(mkdtemp())
ape.config.PROJECT_FOLDER = Path(mkdtemp())


@pytest.fixture(scope="session")
def config():
    yield ape.config


@pytest.fixture(scope="session")
def data_folder(config):
    yield config.DATA_FOLDER


@pytest.fixture(scope="session")
def plugin_manager():
    yield ape.plugin_manager


@pytest.fixture(scope="session")
def accounts():
    yield ape.accounts


@pytest.fixture(scope="session")
def compilers():
    yield ape.compilers


@pytest.fixture(scope="session")
def networks():
    yield ape.networks


@pytest.fixture(scope="session")
def project_folder(config):
    yield config.PROJECT_FOLDER


@pytest.fixture(scope="session")
def project(config):
    yield ape.Project(config.PROJECT_FOLDER)



from typing import Dict, Iterator, List, Type

from dataclassy import dataclass
from pluggy import PluginManager  # type: ignore

from ape.api.accounts import AccountAPI, AccountContainerAPI
from ape.types import AddressType
from ape.utils import cached_property, singledispatchmethod

from .config import ConfigManager
from .converters import ConversionManager
from .networks import NetworkManager


@dataclass
class AccountManager:
    """
    Accounts is a container of containers for AccountAPI objects
    All containers must subclass AccountContainerAPI, and are treated as singletons
    """

    config: ConfigManager
    converters: ConversionManager
    plugin_manager: PluginManager
    network_manager: NetworkManager

    @cached_property
    def containers(self) -> Dict[str, AccountContainerAPI]:
        containers = dict()
        data_folder = self.config.DATA_FOLDER
        data_folder.mkdir(exist_ok=True)
        for plugin_name, (container_type, account_type) in self.plugin_manager.account_types:
            accounts_folder = data_folder / plugin_name
            accounts_folder.mkdir(exist_ok=True)
            containers[plugin_name] = container_type(accounts_folder, account_type)

        return containers

    @property
    def aliases(self) -> Iterator[str]:
        for container in self.containers.values():
            yield from container.aliases

    def get_accounts_by_type(self, type_: Type[AccountAPI]) -> List[AccountAPI]:
        accounts_with_type = []
        for account in self:
            if isinstance(account, type_):
                self._inject_provider(account)
                accounts_with_type.append(account)

        return accounts_with_type

    def __len__(self) -> int:
        return sum(len(container) for container in self.containers.values())

    def __iter__(self) -> Iterator[AccountAPI]:
        for container in self.containers.values():
            for account in container:
                self._inject_provider(account)
                yield account

    def __repr__(self) -> str:
        return "[" + ", ".join(repr(a) for a in self) + "]"

    def load(self, alias: str) -> AccountAPI:
        if alias == "":
            raise ValueError("Cannot use empty string as alias!")

        for account in self:
            if account.alias and account.alias == alias:
                self._inject_provider(account)
                return account

        raise IndexError(f"No account with alias `{alias}`.")

    @singledispatchmethod
    def __getitem__(self, account_id) -> AccountAPI:
        raise NotImplementedError("Cannot use " + type(account_id) + " as account id")

    @__getitem__.register
    def __getitem_int(self, account_id: int) -> AccountAPI:
        for idx, account in enumerate(self.__iter__()):
            if account_id == idx:
                self._inject_provider(account)
                return account

        raise IndexError(f"No account at index `{account_id}`.")

    @__getitem__.register
    def __getitem_str(self, account_str: str) -> AccountAPI:
        account_id = self.converters.convert(account_str, AddressType)

        for container in self.containers.values():
            if account_id in container:
                account = container[account_id]
                self._inject_provider(account)
                return account

        raise IndexError(f"No account with address `{account_id}`.")

    def __contains__(self, address: AddressType) -> bool:
        return any(address in container for container in self.containers.values())

    def _inject_provider(self, account: AccountAPI):
        if self.network_manager.active_provider is not None:
            account.provider = self.network_manager.active_provider



from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Type

from pluggy import PluginManager  # type: ignore

from ape.types import ABI, AddressType
from ape.utils import cached_property

from ..exceptions import NetworkError, NetworkNotFoundError
from .base import abstractdataclass, abstractmethod, dataclass
from .config import ConfigItem

if TYPE_CHECKING:
    from ape.managers.config import ConfigManager
    from ape.managers.networks import NetworkManager

    from .contracts import ContractLog
    from .explorers import ExplorerAPI
    from .providers import ProviderAPI, ReceiptAPI, TransactionAPI


@abstractdataclass
class EcosystemAPI:
    """
    An Ecosystem is a set of related Networks
    """

    name: str  # Set as plugin name
    network_manager: "NetworkManager"
    config_manager: "ConfigManager"
    plugin_manager: PluginManager
    data_folder: Path
    request_header: str

    transaction_class: Type["TransactionAPI"]
    receipt_class: Type["ReceiptAPI"]

    _default_network: str = "development"

    @cached_property
    def config(self) -> ConfigItem:
        return self.config_manager.get_config(self.name)

    @cached_property
    def networks(self) -> Dict[str, "NetworkAPI"]:
        networks = {}
        for _, (ecosystem_name, network_name, network_class) in self.plugin_manager.networks:
            if ecosystem_name == self.name:
                network_folder = self.data_folder / network_name

                networks[network_name] = network_class(
                    name=network_name,
                    ecosystem=self,
                    config_manager=self.config_manager,
                    plugin_manager=self.plugin_manager,
                    data_folder=network_folder,
                    request_header=self.request_header,
                )

        if len(networks) > 0:
            return networks

        else:
            raise NetworkError("No networks found")

    def __post_init__(self):
        if len(self.networks) == 0:
            raise NetworkError("Must define at least one network in ecosystem")

    def __iter__(self) -> Iterator[str]:
        """
        Provides the set of all valid Network names in the ecosystem
        """
        yield from self.networks

    def __getitem__(self, network_name: str) -> "NetworkAPI":
        return self._try_get_network(network_name)

    def __getattr__(self, network_name: str) -> "NetworkAPI":
        return self._try_get_network(network_name)

    def add_network(self, network_name: str, network: "NetworkAPI"):
        """
        Used to attach new networks to an ecosystem (e.g. L2 networks like Optimism)
        """
        if network_name in self.networks:
            raise NetworkError(f"Unable to overwrite existing network '{network_name}'")
        else:
            self.networks[network_name] = network

    @property
    def default_network(self) -> str:
        return self._default_network

    def set_default_network(self, network_name: str):
        if network_name in self.networks:
            self._default_network = network_name
        else:
            message = f"'{network_name}' is not a valid network for ecosystem '{self.name}'"
            raise NetworkError(message)

    @abstractmethod
    def encode_deployment(
        self, deployment_bytecode: bytes, abi: Optional[ABI], *args, **kwargs
    ) -> "TransactionAPI":
        ...

    @abstractmethod
    def encode_transaction(
        self, address: AddressType, abi: ABI, *args, **kwargs
    ) -> "TransactionAPI":
        ...

    @abstractmethod
    def decode_event(self, abi: ABI, receipt: "ReceiptAPI") -> "ContractLog":
        ...

    def _try_get_network(self, network_name):
        if network_name in self.networks:
            return self.networks[network_name]
        else:
            raise NetworkNotFoundError(network_name)


class ProviderContextManager:
    # NOTE: Class variable, so it will manage stack across instances of this object
    _connected_providers: List["ProviderAPI"] = []

    def __init__(self, network_manager: "NetworkManager", provider: "ProviderAPI"):
        self.network_manager = network_manager
        self.provider = provider

    def __enter__(self, *args, **kwargs):
        # If we are already connected to a provider, disconnect and add
        # it to our stack of providers that were connected
        if self._connected_providers:
            self._connected_providers[-1].disconnect()

        # Connect to our provider
        self.provider.connect()
        self.network_manager.active_provider = self.provider
        self._connected_providers.append(self.provider)

        return self.provider

    def __exit__(self, *args, **kwargs):
        # Put our providers back the way it was
        provider = self._connected_providers.pop()
        if self.provider != provider:
            raise ValueError("Previous provider value unknown")

        provider.disconnect()

        if self._connected_providers:
            self._connected_providers[-1].connect()
            self.network_manager.active_provider = self._connected_providers[-1]


@dataclass
class NetworkAPI:
    """
    A Network is a wrapper around a Provider for a specific Ecosystem.
    """

    name: str  # Name given when registered in ecosystem
    ecosystem: EcosystemAPI
    config_manager: "ConfigManager"
    plugin_manager: PluginManager
    data_folder: Path  # For caching any data that might need caching
    request_header: str

    _default_provider: str = ""

    @cached_property
    def config(self) -> ConfigItem:
        return self.config_manager.get_config(self.ecosystem.name)

    @property
    def chain_id(self) -> int:
        # NOTE: Unless overridden, returns same as `provider.chain_id`
        provider = self.ecosystem.network_manager.active_provider

        if not provider:
            message = (
                "Cannot determine `chain_id`, please make sure you are connected to a provider"
            )
            raise NetworkError(message)

        return provider.chain_id

    @property
    def network_id(self) -> int:
        # NOTE: Unless overridden, returns same as chain_id
        return self.chain_id

    @cached_property
    def explorer(self) -> Optional["ExplorerAPI"]:
        for plugin_name, plugin_tuple in self.plugin_manager.explorers:
            ecosystem_name, network_name, explorer_class = plugin_tuple

            if self.ecosystem.name == ecosystem_name and self.name == network_name:
                # Return the first registered explorer (skipping any others)
                return explorer_class(
                    name=plugin_name,
                    network=self,
                    request_header=self.request_header,
                )

        return None  # May not have an block explorer

    @cached_property
    def providers(self):  # -> Dict[str, Partial[ProviderAPI]]
        providers = {}

        for plugin_name, plugin_tuple in self.plugin_manager.providers:
            ecosystem_name, network_name, provider_class = plugin_tuple

            if self.ecosystem.name == ecosystem_name and self.name == network_name:
                # NOTE: Lazily load and provider config on load
                providers[plugin_name] = partial(
                    provider_class,
                    name=plugin_name,
                    config=self.config_manager.get_config(plugin_name),
                    network=self,
                    # NOTE: No need to have separate folder, caching should be interoperable
                    data_folder=self.data_folder,
                    request_header=self.request_header,
                )

        if len(providers) > 0:
            return providers

        else:
            raise NetworkError("No network providers found")

    def use_provider(
        self,
        provider_name: str,
        provider_settings: dict = None,
    ) -> ProviderContextManager:
        if provider_settings is None:
            provider_settings = {}

        if ":" in provider_name:
            # NOTE: Shortcut that allows `--network ecosystem:network:http://...` to work
            provider_settings["uri"] = provider_name
            provider_name = provider_name.split(":")[0]

        if provider_name in self.providers:
            return ProviderContextManager(
                self.ecosystem.network_manager,
                self.providers[provider_name](provider_settings=provider_settings),
            )

        else:
            message = (
                f"'{provider_name}' is not a valid network for ecosystem '{self.ecosystem.name}'"
            )
            raise NetworkError(message)

    @property
    def default_provider(self) -> str:
        return self._default_provider or list(self.providers)[0]

    def set_default_provider(self, provider_name: str):
        if provider_name in self.providers:
            self._default_provider = provider_name
        else:
            raise NetworkError(f"No providers found for network '{self.name}'")

    def use_default_provider(self) -> ProviderContextManager:
        # NOTE: If multiple providers, use whatever is "first" registered
        return self.use_provider(self.default_provider)


def create_network_type(chain_id: int, network_id: int) -> Type[NetworkAPI]:
    """
    Helper function that allows creating a :class:`NetworkAPI` subclass easily.
    """

    class network_def(NetworkAPI):
        @property
        def chain_id(self) -> int:
            return chain_id

        @property
        def network_id(self) -> int:
            return network_id

    return network_def



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from eth_utils import to_bytes

from ape.logging import logger
from ape.types import ABI, AddressType, ContractType

from ..exceptions import ContractCallError, ContractDeployError
from .address import Address, AddressAPI
from .base import dataclass
from .providers import ProviderAPI, ReceiptAPI, TransactionAPI

if TYPE_CHECKING:
    from ape.managers.converters import ConversionManager
    from ape.managers.networks import NetworkManager


@dataclass
class ContractConstructor:
    deployment_bytecode: bytes
    abi: Optional[ABI]
    provider: ProviderAPI

    def __post_init__(self):
        if len(self.deployment_bytecode) == 0:
            raise ContractDeployError("No bytecode to deploy")

    def __repr__(self) -> str:
        return self.abi.signature if self.abi else "constructor()"

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_deployment(
            self.deployment_bytecode, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> ReceiptAPI:
        if "sender" in kwargs:
            sender = kwargs["sender"]
            kwargs["sender"] = sender.address

            txn = self.encode(*args, **kwargs)
            return sender.call(txn)

        else:
            txn = self.encode(*args, **kwargs)
            return self.provider.send_transaction(txn)


@dataclass
class ContractCall:
    abi: ABI
    address: AddressType
    provider: ProviderAPI

    def __repr__(self) -> str:
        return self.abi.signature

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> Any:
        txn = self.encode(*args, **kwargs)

        if "sender" in kwargs and not isinstance(kwargs["sender"], str):
            txn.sender = kwargs["sender"].address

        raw_output = self.provider.send_call(txn)
        tuple_output = self.provider.network.ecosystem.decode_calldata(  # type: ignore
            self.abi,
            raw_output,
        )

        # NOTE: Returns a tuple, so make sure to handle all the cases
        if len(tuple_output) < 2:
            return tuple_output[0] if len(tuple_output) == 1 else None

        else:
            # TODO: Handle struct output
            return tuple_output


@dataclass
class ContractCallHandler:
    provider: ProviderAPI
    address: AddressType
    abis: List[ABI]

    def __repr__(self) -> str:
        abis = sorted(self.abis, key=lambda abi: len(abi.inputs))
        return abis[-1].signature

    def __call__(self, *args, **kwargs) -> Any:
        selected_abi = _select_abi(self.abis, args)
        if not selected_abi:
            raise ContractCallError()

        return ContractCall(  # type: ignore
            abi=selected_abi,
            address=self.address,
            provider=self.provider,
        )(*args, **kwargs)


def _select_abi(abis, args):
    selected_abi = None
    for abi in abis:
        if len(args) == len(abi.inputs):
            selected_abi = abi

    return selected_abi


@dataclass
class ContractTransaction:
    abi: ABI
    address: AddressType
    provider: ProviderAPI

    def __repr__(self) -> str:
        return self.abi.signature

    def encode(self, *args, **kwargs) -> TransactionAPI:
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *args, **kwargs
        )

    def __call__(self, *args, **kwargs) -> ReceiptAPI:

        if "sender" in kwargs:
            sender = kwargs["sender"]
            kwargs["sender"] = sender.address

            txn = self.encode(*args, **kwargs)
            return sender.call(txn)

        else:
            raise ContractCallError("Must specify a `sender`")


@dataclass
class ContractTransactionHandler:
    provider: ProviderAPI
    address: AddressType
    abis: List[ABI]

    def __repr__(self) -> str:
        abis = sorted(self.abis, key=lambda abi: len(abi.inputs))
        return abis[-1].signature

    def __call__(self, *args, **kwargs) -> ReceiptAPI:
        selected_abi = _select_abi(self.abis, args)
        if not selected_abi:
            raise ContractCallError()

        return ContractTransaction(  # type: ignore
            abi=selected_abi,
            address=self.address,
            provider=self.provider,
        )(*args, **kwargs)


@dataclass
class ContractLog:
    name: str
    data: Dict[str, Any]


@dataclass
class ContractEvent:
    provider: ProviderAPI
    address: str
    abis: List[ABI]
    cached_logs: List[ContractLog] = []


class ContractInstance(AddressAPI):
    _address: AddressType
    _contract_type: ContractType

    def __repr__(self) -> str:
        return f"<{self._contract_type.contractName} {self.address}>"

    @property
    def address(self) -> AddressType:
        return self._address

    def __dir__(self) -> List[str]:
        # This displays methods to IPython on `c.[TAB]` tab completion
        return list(super(AddressAPI, self).__dir__()) + [
            abi.name for abi in self._contract_type.abi
        ]

    def __getattr__(self, attr_name: str) -> Any:
        handlers = {
            "events": ContractEvent,
            "calls": ContractCallHandler,
            "transactions": ContractTransactionHandler,
        }

        def get_handler(abi_type: str) -> Any:
            selected_abis = [
                abi for abi in getattr(self._contract_type, abi_type) if abi.name == attr_name
            ]

            if not selected_abis:
                return  # No ABIs found for this type

            kwargs = {
                "provider": self.provider,
                "address": self.address,
                "abis": selected_abis,
            }

            try:
                return handlers[abi_type](**kwargs)  # type: ignore

            except Exception as e:
                # NOTE: Just a hack, because `__getattr__` *must* raise `AttributeError`
                raise AttributeError from e

        # Reverse search for the proper handler for this ABI name, if one exists
        for abi_type in handlers:
            handler = get_handler(abi_type)
            if handler:
                return handler
            # else: No ABI found with `attr_name`

        # No ABIs w/ name `attr_name` found at all
        raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")


@dataclass
class ContractContainer:
    _provider: ProviderAPI
    _contract_type: ContractType

    def __repr__(self) -> str:
        return f"<{self._contract_type.contractName}>"

    def at(self, address: str) -> ContractInstance:
        return ContractInstance(  # type: ignore
            _address=address,
            _provider=self._provider,
            _contract_type=self._contract_type,
        )

    @property
    def _deployment_bytecode(self) -> bytes:
        if (
            self._contract_type.deploymentBytecode
            and self._contract_type.deploymentBytecode.bytecode
        ):
            return to_bytes(hexstr=self._contract_type.deploymentBytecode.bytecode)

        else:
            return b""

    @property
    def _runtime_bytecode(self) -> bytes:
        if self._contract_type.runtimeBytecode and self._contract_type.runtimeBytecode.bytecode:
            return to_bytes(hexstr=self._contract_type.runtimeBytecode.bytecode)

        else:
            return b""

    def __call__(self, *args, **kwargs) -> TransactionAPI:
        constructor = ContractConstructor(  # type: ignore
            abi=self._contract_type.constructor,
            provider=self._provider,
            deployment_bytecode=self._deployment_bytecode,
        )
        return constructor.encode(*args, **kwargs)


def _Contract(
    address: Union[str, AddressAPI, AddressType],
    networks: "NetworkManager",
    converters: "ConversionManager",
    contract_type: Optional[ContractType] = None,
) -> AddressAPI:
    """
    Function used to triage whether we have a contract type available for
    the given address/network combo, or explicitly provided. If none are found,
    returns a simple ``Address`` instance instead of throwing (provides a warning)
    """

    # Check contract cache (e.g. previously deployed/downloaded contracts)
    # TODO: Add ``contract_cache`` dict-like object to ``NetworkAPI``
    # network = provider.network
    # if not contract_type and address in network.contract_cache:
    #    contract_type = network.contract_cache[address]

    # Check explorer API/cache (e.g. publicly published contracts)
    # TODO: Add ``get_contract_type`` to ``ExplorerAPI``
    # TODO: Store in ``NetworkAPI.contract_cache`` to reduce API calls
    # explorer = provider.network.explorer
    # if not contract_type and explorer:
    #    contract_type = explorer.get_contract_type(address)

    # We have a contract type either:
    #   1) explicity provided,
    #   2) from network cache, or
    #   3) from explorer
    if contract_type:
        return ContractInstance(  # type: ignore
            _address=converters.convert(address, AddressType),
            _provider=networks.active_provider,
            _contract_type=contract_type,
        )

    else:
        # We don't have a contract type from any source, provide raw address instead
        logger.warning(f"No contract type found for {address}")
        return Address(  # type: ignore
            _address=converters.convert(address, AddressType),
            _provider=networks.active_provider,
        )



import pytest  # type: ignore

from ape import __all__


# NOTE: We export `__all__` into the IPython session that the console runs in
@pytest.mark.parametrize("item", __all__)
def test_console(ape_cli, runner, item):
    result = runner.invoke(ape_cli, "console", input=f"{item}\nexit\n")
    assert result.exit_code == 0



from typing import Type

from ape.api.config import ConfigItem

from .pluggy_patch import PluginType, hookspec


class Config(PluginType):
    @hookspec
    def config_class(self) -> Type[ConfigItem]:
        """
        Returns a :class:`ape.api.config.ConfigItem` parser class that can be
        used to deconstruct the user config options for this plugins.

        NOTE: If none are specified, all injected :class:`ape.api.config.ConfigItem`'s are empty.
        """



from typing import Any, List, Optional, Type

import click
from click import Context, Parameter

from ape import accounts
from ape.api.accounts import AccountAPI


class Alias(click.Choice):
    """Wraps ``click.Choice`` to load account aliases for the active project at runtime.

    Provide an ``account_type`` to limit the type of account to choose from.
    Defaults to all account types in ``choices()``.
    """

    name = "alias"

    def __init__(self, account_type: Optional[Type[AccountAPI]] = None):
        # NOTE: we purposely skip the constructor of `Choice`
        self.case_sensitive = False
        self._account_type = account_type

    @property
    def choices(self) -> List[str]:  # type: ignore
        options = (
            list(accounts)
            if not self._account_type
            else accounts.get_accounts_by_type(self._account_type)
        )
        return [a.alias for a in options if a.alias is not None]


class PromptChoice(click.ParamType):
    """
    A choice option or argument from user selection.
    """

    def __init__(self, choices):
        self.choices = choices
        self.choice_index = None

    def print_choices(self):
        choices = dict(enumerate(self.choices, 1))
        for choice in choices:
            click.echo(f"{choice}. {choices[choice]}")
        click.echo()

    def convert(
        self, value: Any, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Optional[str]:
        # noinspection PyBroadException
        try:
            self.choice_index = int(value) - 1
            if self.choice_index < 0:
                self.fail("Invalid choice", param=param)

            choice = self.choices[self.choice_index]
            return choice
        except Exception:
            self.fail("Invalid choice", param=param)



import click

from ape.logging import logger


class Abort(click.ClickException):
    """Wrapper around a CLI exception"""

    def show(self, file=None):
        """Override default ``show`` to print CLI errors in red text."""
        logger.error(self.format_message())



from itertools import chain
from pathlib import Path
from typing import Dict, List, Optional

import click

from ape.cli import ape_cli_context
from ape.types import ContractType

flatten = chain.from_iterable


class _ContractsSource:
    """
    A helper class that is able to figure out which source files to
    compile.
    """

    def __init__(self, use_cache: bool):
        # NOTE: Lazy load so that testing works properly
        from ape import project

        self._project = project
        self._use_cache = use_cache

    @property
    def root(self) -> Path:
        return self._project.path / "contracts"

    def select_paths(self, file_paths: Optional[List[Path]]):
        # If not given sources, assume user wants to compile all source files.
        # Excludes files without registered compilers.
        if not file_paths:
            return self._project.sources

        expanded_file_paths = flatten([d.rglob("*.*") if d.is_dir() else [d] for d in file_paths])
        return [c.resolve() for c in expanded_file_paths if c.resolve() in self._project.sources]

    def compile(self) -> Dict[str, ContractType]:
        # TODO: only compile selected contracts
        return self._project.load_contracts(self._use_cache)


@click.command(short_help="Compile select contract source files")
@click.argument(
    "filepaths",
    nargs=-1,
    type=click.Path(exists=True, path_type=Path),
)
@click.option(
    "-f",
    "--force",
    "use_cache",
    flag_value=False,
    default=True,
    is_flag=True,
    help="Force recompiling selected contracts",
)
@click.option(
    "-s",
    "--size",
    "display_size",
    default=False,
    is_flag=True,
    help="Show deployment bytecode size for all contracts",
)
@ape_cli_context()
def cli(cli_ctx, filepaths, use_cache, display_size):
    """
    Compiles the manifest for this project and saves the results
    back to the manifest.

    Note that ape automatically recompiles any changed contracts each time
    a project is loaded. You do not have to manually trigger a recompile.
    """
    source = _ContractsSource(use_cache)
    missing_source = not source.root.exists() or not source.root.iterdir()
    if not filepaths and missing_source:
        cli_ctx.logger.warning("No 'contracts/' directory detected")
        return

    selected_file_paths = source.select_paths(filepaths)
    source_file_paths = list(source.root.iterdir())
    unable_to_select_all = len(source_file_paths) > len(selected_file_paths)

    if not selected_file_paths or unable_to_select_all:
        _warn_for_missing_extensions(cli_ctx, selected_file_paths, source_file_paths)

    contract_types = source.compile()

    if display_size:
        _display_byte_code_sizes(cli_ctx, contract_types)


def _warn_for_missing_extensions(cli_ctx, registered_sources: List[Path], all_sources: List[Path]):
    """
    Figures out what extensions are missing from registered compilers and warns
    the user about them.
    """
    extensions_unable_to_compile = set()
    for path in all_sources:
        if path not in registered_sources:
            extensions_unable_to_compile.add(path.suffix)

    if extensions_unable_to_compile:
        extensions_str = ", ".join(extensions_unable_to_compile)
        message = f"No compilers detected for the following extensions: {extensions_str}"
    else:
        message = "Nothing to compile"

    cli_ctx.logger.warning(message)


def _display_byte_code_sizes(cli_ctx, contract_types: Dict[str, ContractType]):
    # Display bytecode size for *all* contract types (not just ones we compiled)
    code_size = []
    for contract in contract_types.values():
        if not contract.deploymentBytecode:
            continue  # Skip if not bytecode to display

        bytecode = contract.deploymentBytecode.bytecode

        if bytecode:
            code_size.append((contract.contractName, len(bytecode) // 2))

    if not code_size:
        cli_ctx.logger.info("No contracts with bytecode to display")
        return

    click.echo()
    click.echo("============ Deployment Bytecode Sizes ============")
    indent = max(len(i[0]) for i in code_size)
    for name, size in sorted(code_size, key=lambda k: k[1], reverse=True):
        pct = size / 24577
        # pct_color = color(next((i[1] for i in CODESIZE_COLORS if pct >= i[0]), ""))
        # TODO Get colors fixed for bytecode size output
        # click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct_color}{pct:.2%}{color})")
        click.echo(f"  {name:<{indent}}  -  {size:>6,}B  ({pct:.2%})")

    click.echo()



from typing import Tuple, Type

from ape.api.accounts import AccountAPI, AccountContainerAPI

from .pluggy_patch import PluginType, hookspec


class AccountPlugin(PluginType):
    @hookspec
    def account_types(self) -> Tuple[Type[AccountContainerAPI], Type[AccountAPI]]:
        """
        Must return a tuple of an Account Container and Account type
        """



import json
from pathlib import Path
from typing import Dict, List

from dataclassy import dataclass

from ape.api.config import ConfigDict, ConfigItem
from ape.exceptions import ConfigError
from ape.plugins import PluginManager
from ape.utils import load_config

CONFIG_FILE_NAME = "ape-config.yaml"


@dataclass
class ConfigManager:
    DATA_FOLDER: Path
    REQUEST_HEADER: Dict
    PROJECT_FOLDER: Path
    name: str = ""
    version: str = ""
    dependencies: List[str] = []
    plugin_manager: PluginManager
    _plugin_configs: Dict[str, ConfigItem] = dict()

    def __post_init__(self):
        config_file = self.PROJECT_FOLDER / CONFIG_FILE_NAME

        if config_file.exists():
            user_config = load_config(config_file)
        else:
            user_config = {}

        # Top level config items
        self.name = user_config.pop("name", "")
        self.version = user_config.pop("version", "")
        self.dependencies = user_config.pop("dependencies", [])

        for plugin_name, config_class in self.plugin_manager.config_class:
            # NOTE: `dict.pop()` is used for checking if all config was processed
            user_override = user_config.pop(plugin_name, {})

            if config_class != ConfigDict:
                # NOTE: Will raise if improperly provided keys
                config = config_class(**user_override)

                # NOTE: Should raise if settings violate some sort of plugin requirement
                config.validate_config()

            else:
                # NOTE: Just use it directly as a dict if `ConfigDict` is passed
                config = user_override

            self._plugin_configs[plugin_name] = config

        if len(user_config.keys()) > 0:
            raise ConfigError("Unprocessed config items")

    def get_config(self, plugin_name: str) -> ConfigItem:
        if plugin_name not in self._plugin_configs:
            # plugin has no registered config class, so return empty config
            return ConfigItem()

        return self._plugin_configs[plugin_name]

    def serialize(self) -> Dict:
        project_config = dict()

        for name, config in self._plugin_configs.items():
            # NOTE: `config` is either `ConfigItem` or `dict`
            project_config[name] = config.serialize() if isinstance(config, ConfigItem) else config

        return project_config

    def __str__(self) -> str:
        return json.dumps(self.serialize())



from pathlib import Path
from typing import List, Set

from ape.types import ContractType

from .base import abstractdataclass, abstractmethod


@abstractdataclass
class CompilerAPI:
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def get_versions(self, all_paths: List[Path]) -> Set[str]:
        ...

    @abstractmethod
    def compile(self, contract_filepaths: List[Path]) -> List[ContractType]:
        """
        Compile the source given ``pkg_manifest``.
        All compiler plugins must implement this function.
        """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self) -> str:
        return f"{self.name}>"



from pathlib import Path
from typing import Callable, Iterator, List, Optional, Type, Union

from ape.types import (
    AddressType,
    ContractType,
    MessageSignature,
    SignableMessage,
    TransactionSignature,
)
from ape.utils import cached_property

from ..exceptions import AccountsError, AliasAlreadyInUseError
from .address import AddressAPI
from .base import abstractdataclass, abstractmethod
from .contracts import ContractContainer, ContractInstance
from .providers import ReceiptAPI, TransactionAPI


# NOTE: AddressAPI is a dataclass already
class AccountAPI(AddressAPI):
    container: "AccountContainerAPI"

    def __dir__(self) -> List[str]:
        # This displays methods to IPython on `a.[TAB]` tab completion
        return list(super(AddressAPI, self).__dir__()) + [
            "alias",
            "sign_message",
            "sign_transaction",
            "call",
            "transfer",
            "deploy",
        ]

    @property
    def alias(self) -> Optional[str]:
        """
        Override with whatever alias might want to use, if applicable
        """
        return None

    @abstractmethod
    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        ...

    @abstractmethod
    def sign_transaction(self, txn: TransactionAPI) -> Optional[TransactionSignature]:
        ...

    def call(self, txn: TransactionAPI, send_everything: bool = False) -> ReceiptAPI:
        # NOTE: Use "expected value" for Chain ID, so if it doesn't match actual, we raise
        txn.chain_id = self.provider.network.chain_id

        # NOTE: Allow overriding nonce, assume user understand what this does
        if txn.nonce is None:
            txn.nonce = self.nonce
        elif txn.nonce < self.nonce:
            raise AccountsError("Invalid nonce, will not publish")

        # TODO: Add `GasEstimationAPI`
        if txn.gas_price is None:
            txn.gas_price = self.provider.gas_price
        # else: assume user specified a correct price, or will take too much time to confirm

        # NOTE: Allow overriding gas limit
        if txn.gas_limit is None:
            txn.gas_limit = 0  # NOTE: Need a starting estimate
            txn.gas_limit = self.provider.estimate_gas_cost(txn)
        # else: assume user specified the correct amount or txn will fail and waste gas

        if send_everything:
            txn.value = self.balance - txn.gas_limit * txn.gas_price

        if txn.gas_limit * txn.gas_price + txn.value > self.balance:
            raise AccountsError("Transfer value meets or exceeds account balance")

        txn.signature = self.sign_transaction(txn)

        if not txn.signature:
            raise AccountsError("The transaction was not signed")

        return self.provider.send_transaction(txn)

    @cached_property
    def _convert(self) -> Callable:
        # NOTE: Need to differ loading this property
        from ape import convert

        return convert

    def transfer(
        self,
        account: Union[str, AddressType, "AddressAPI"],
        value: Union[str, int, None] = None,
        data: Union[bytes, str, None] = None,
        **kwargs,
    ) -> ReceiptAPI:
        txn = self._transaction_class(  # type: ignore
            sender=self.address,
            receiver=self._convert(account, AddressType),
            **kwargs,
        )

        if data:
            txn.data = self._convert(data, bytes)

        if value:
            txn.value = self._convert(value, int)

        return self.call(txn, send_everything=value is None)

    def deploy(self, contract_type: ContractType, *args, **kwargs) -> ContractInstance:
        c = ContractContainer(  # type: ignore
            _provider=self.provider,
            _contract_type=contract_type,
        )

        txn = c(*args, **kwargs)
        txn.sender = self.address
        receipt = self.call(txn)

        if not receipt.contract_address:
            raise AccountsError(f"'{receipt.txn_hash}' did not create a contract")

        return ContractInstance(  # type: ignore
            _provider=self.provider,
            _address=receipt.contract_address,
            _contract_type=contract_type,
        )


@abstractdataclass
class AccountContainerAPI:
    data_folder: Path
    account_type: Type[AccountAPI]

    @property
    @abstractmethod
    def aliases(self) -> Iterator[str]:
        ...

    @abstractmethod
    def __len__(self) -> int:
        ...

    @abstractmethod
    def __iter__(self) -> Iterator[AccountAPI]:
        ...

    def __getitem__(self, address: AddressType) -> AccountAPI:
        for account in self.__iter__():
            if account.address == address:
                return account

        raise IndexError(f"No local account {address}.")

    def append(self, account: AccountAPI):
        self._verify_account_type(account)

        if account.address in self:
            raise AccountsError(f"Account '{account.address}' already in container")

        self._verify_unused_alias(account)

        self.__setitem__(account.address, account)

    def __setitem__(self, address: AddressType, account: AccountAPI):
        raise NotImplementedError("Must define this method to use `container.append(acct)`")

    def remove(self, account: AccountAPI):
        self._verify_account_type(account)

        if account.address not in self:
            raise AccountsError(f"Account '{account.address}' not known")

        self.__delitem__(account.address)

    def __delitem__(self, address: AddressType):
        raise NotImplementedError("Must define this method to use `container.remove(acct)`")

    def __contains__(self, address: AddressType) -> bool:
        try:
            self.__getitem__(address)
            return True

        except IndexError:
            return False

    def _verify_account_type(self, account):
        if not isinstance(account, self.account_type):
            message = (
                f"Container '{type(account).__name__}' is an incorrect "
                f"type for container '{type(self.account_type).__name__}'"
            )
            raise AccountsError(message)

    def _verify_unused_alias(self, account):
        if account.alias and account.alias in self.aliases:
            raise AliasAlreadyInUseError(account.alias)



"""
Used with the `docs` Github action to make versioned docs directories in the
gh-pages branch.
"""
import argparse
import os
import re
import shutil
import subprocess
from distutils.version import LooseVersion
from pathlib import Path

REDIRECT_HTML = """
<!DOCTYPE html>
<meta charset="utf-8">
<title>Redirecting...</title>
<meta http-equiv="refresh" content="0; URL=./stable/">
"""

DOCS_BUILD_PATH = Path("docs/_build")


def log(*args, **kwargs):
    print(*args, **kwargs)  # noqa: T001


def run(args):
    log("Running:", args)
    subprocess.check_call(args)


def main():
    # clean out the contents of the build directory, but leave the directory in
    # place so that serving up files continues to work when you're developing
    # docs locally
    DOCS_BUILD_PATH.mkdir(exist_ok=True)
    for path in DOCS_BUILD_PATH.iterdir():
        shutil.rmtree(path) if path.is_dir() else path.unlink()

    if "pull_request" in os.environ.get("GITHUB_EVENT_NAME", ""):
        # build only the current working dir's docs for a PR because PR builds
        # don't have all the refs needed for multiversion
        run(["sphinx-build", "docs", "docs/_build"])
        return  # none of the steps below need to run if we're building for a PR
    else:
        # build docs for each version
        run(["sphinx-multiversion", "docs", "docs/_build"])

    # move the current branch to "latest"
    branch = subprocess.check_output(["git", "branch", "--show-current"]).decode("ascii").strip()
    branch = branch or "main"
    if (DOCS_BUILD_PATH / branch).is_dir():
        Path(DOCS_BUILD_PATH / branch).rename(DOCS_BUILD_PATH / "latest")

    # clean up static files so we don't need to host the same 10+ MB of web
    # fonts for each version of the docs
    for d in Path("docs/_build").glob("**/fonts"):
        if "latest" in str(d):
            continue  # leave only the copy of the static files from the latest version
        shutil.rmtree(d)

    # copy the highest released version to "stable"
    all_releases = [
        LooseVersion(d.name) for d in DOCS_BUILD_PATH.iterdir() if re.match(r"v\d+\.", d.name)
    ]
    no_pre_releases = [v for v in all_releases if "-" not in str(v)]
    stable = None
    if no_pre_releases:
        stable = max(no_pre_releases)
    elif all_releases:
        stable = max(all_releases)
    else:
        log("WARNING: Couldn't find any released versions. Going to use 'latest' for 'stable'.")
        stable = "latest"
    log(f"Copying latest stable release {stable} to 'stable'.")
    shutil.copytree(DOCS_BUILD_PATH / str(stable), DOCS_BUILD_PATH / "stable")

    # set up the redirect at /index.html
    with open(DOCS_BUILD_PATH / "index.html", "w") as f:
        f.write(REDIRECT_HTML)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--rsync",
        help="Where to put the fully built docs subdir for serving in development. "
        "/tmp/projectname is recommended.",
    )
    args = parser.parse_args()

    main()

    if args.rsync:
        run(["rsync", "-pthrvz", "--delete", "./docs/_build/", args.rsync])
        log("\n")
        log(
            (
                "NOTE: To serve these files in development, run `python3 -m http.server` inside "
                "`{}`, then go to http://127.0.0.1:8000/projectname in the browser."
            ).format(Path(args.rsync).parent)
        )
        log(
            "NOTE: If you're making changes to docs locally, go to the 'latest' branch to see "
            "your changes. Also, due to the way sphinx-multiversion integrates with git, you need "
            "to commit your changes each time before building."
        )
        log()



from ape.cli.choices import Alias, PromptChoice
from ape.cli.options import (
    ape_cli_context,
    existing_alias_argument,
    network_option,
    non_existing_alias_argument,
    skip_confirmation_option,
    verbose_option,
)
from ape.cli.utils import Abort

__all__ = [
    "Abort",
    "Alias",
    "ape_cli_context",
    "existing_alias_argument",
    "network_option",
    "non_existing_alias_argument",
    "PromptChoice",
    "skip_confirmation_option",
    "verbose_option",
]



from typing import Iterator, Tuple, Type

from ape.api import ConverterAPI

from .pluggy_patch import PluginType, hookspec


class ConversionPlugin(PluginType):
    @hookspec
    def converters(self) -> Iterator[Tuple[str, Type[ConverterAPI]]]:
        """
        Must return an iterator of tuples of a string ABIType and an ConverterAPI subclass
        """



from typing import Iterator, Tuple, Type

from ape.api import EcosystemAPI, ExplorerAPI, NetworkAPI, ProviderAPI

from .pluggy_patch import PluginType, hookspec


class EcosystemPlugin(PluginType):
    @hookspec
    def ecosystems(self) -> Iterator[Type[EcosystemAPI]]:
        """
        Must return an iterator of EcosystemAPI subclasses
        """


class NetworkPlugin(PluginType):
    @hookspec
    def networks(self) -> Iterator[Tuple[str, str, Type[NetworkAPI]]]:
        """
        Must return an iterator of tuples of:
            the target Ecosystem plugin's name
            the Network name
            a NetworkAPI subclass
        """


class ProviderPlugin(PluginType):
    @hookspec
    def providers(self) -> Iterator[Tuple[str, str, Type[ProviderAPI]]]:
        """
        Must return an iterator of tuples of:
            the target Ecosystem plugin's name
            the Network it works with (which must be valid Network in the Ecosystem)
            a ProviderAPI subclass
        """


class ExplorerPlugin(PluginType):
    @hookspec
    def explorers(self) -> Iterator[Tuple[str, str, Type[ExplorerAPI]]]:
        """
        Must return an iterator of tuples of:
            the target Ecosystem plugin's name
            the Network it works with (which must be valid Network in the Ecosystem)
            a ExplorerAPI subclass
        """



from typing import Optional

import click

from ape import networks
from ape.api import EcosystemAPI

_SPACED_TAB = "  "


@click.group(short_help="Manage networks")
def cli():
    """
    Command-line helper for managing networks.
    """


@cli.command(name="list", short_help="List registered networks")
def _list():
    click.echo("ecosystems:")
    for ecosystem_name in networks:
        _echo_ecosystem(ecosystem_name)


def _echo_ecosystem(name: str):
    _output_line(name, default_value=networks.default_ecosystem.name)
    ecosystem = networks[name]

    for network_name in getattr(networks, name):
        _echo_network(ecosystem, network_name)


def _echo_network(ecosystem: EcosystemAPI, name: str):
    _output_line(name, num_of_tabs=1, default_value=ecosystem.default_network)
    network = ecosystem[name]

    if network.explorer:
        click.echo(f"{_SPACED_TAB * 2}explorer: {network.explorer.name}")

    click.echo(f"{_SPACED_TAB * 2}providers:")

    for provider_name in network.providers:
        _output_line(provider_name, num_of_tabs=2, default_value=network.default_provider, key=None)


def _output_line(
    value: str, num_of_tabs: int = 0, default_value: str = None, key: Optional[str] = "name"
):
    key = f"{key}: " if key else ""
    comment = "  # Default" if value == default_value else ""
    output = f"{_SPACED_TAB * num_of_tabs}- {key}{value}{comment}"
    click.echo(output)



from pathlib import Path

import pytest

from ape.utils import get_relative_path

_TEST_DIRECTORY_PATH = Path("/This/is/a/test/")
_TEST_FILE_PATH = _TEST_DIRECTORY_PATH / "scripts" / "script.py"


def test_get_relative_path_from_project():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_DIRECTORY_PATH)
    expected = Path("scripts/script.py")
    assert actual == expected


def test_get_relative_path_given_relative_path():
    relative_script_path = Path("../deploy.py")
    with pytest.raises(ValueError) as err:
        get_relative_path(relative_script_path, _TEST_DIRECTORY_PATH)

    assert str(err.value) == "'target' must be an absolute path"

    relative_project_path = Path("../This/is/a/test")

    with pytest.raises(ValueError) as err:
        get_relative_path(_TEST_FILE_PATH, relative_project_path)

    assert str(err.value) == "'anchor' must be an absolute path"


def test_get_relative_path_same_path():
    actual = get_relative_path(_TEST_FILE_PATH, _TEST_FILE_PATH)
    assert actual == Path()


def test_get_relative_path_roots():
    root = Path("/")
    actual = get_relative_path(root, root)
    assert actual == Path()



# Inspired / borrowed from the `click-logging` python package.
import logging
import sys
from enum import Enum
from typing import IO

import click


class LogLevel(Enum):
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    SUCCESS = logging.INFO + 1


logging.addLevelName(LogLevel.SUCCESS.value, LogLevel.SUCCESS.name)
logging.SUCCESS = LogLevel.SUCCESS.value  # type: ignore


def success(self, message, *args, **kws):
    """This method gets injecting into python's `logging` module
    to handle logging at this level."""
    if self.isEnabledFor(LogLevel.SUCCESS.value):
        # Yes, logger takes its '*args' as 'args'.
        self._log(LogLevel.SUCCESS.value, message, args, **kws)


logging.Logger.success = success  # type: ignore


CLICK_STYLE_KWARGS = {
    LogLevel.ERROR: dict(fg="bright_red"),
    LogLevel.WARNING: dict(fg="bright_red"),
    LogLevel.INFO: dict(fg="blue"),
    LogLevel.SUCCESS: dict(fg="bright_green"),
}
CLICK_ECHO_KWARGS = {
    LogLevel.ERROR: dict(err=True),
    LogLevel.WARNING: dict(err=True),
    LogLevel.INFO: dict(),
    LogLevel.SUCCESS: dict(),
}


# Borrowed from `click._compat`.
def _isatty(stream: IO) -> bool:
    """Returns ``True`` if the stream is part of tty.
    Borrowed from ``click._compat``."""
    # noinspection PyBroadException
    try:
        return stream.isatty()
    except Exception:
        return False


class ApeColorFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(fmt="%(levelname)s: %(message)s")

    def format(self, record):
        if _isatty(sys.stdout) and _isatty(sys.stderr):
            # only color log messages when sys.stdout and sys.stderr are sent to the terminal
            level = LogLevel(record.levelno)
            styles = CLICK_STYLE_KWARGS.get(level, {})
            record.levelname = click.style(record.levelname, **styles)

        return super().format(record)


class ClickHandler(logging.Handler):
    def __init__(self, echo_kwargs):
        super().__init__()
        self.echo_kwargs = echo_kwargs

    def emit(self, record):
        try:
            msg = self.format(record)
            level = record.levelname.lower()
            if self.echo_kwargs.get(level):
                click.echo(msg, **self.echo_kwargs[level])
            else:
                click.echo(msg)
        except Exception:
            self.handleError(record)


def _get_logger(name) -> logging.Logger:
    """Get a logger with the given ``name`` and configure it for usage with Click."""
    cli_logger = logging.getLogger(name)
    handler = ClickHandler(echo_kwargs=CLICK_ECHO_KWARGS)
    handler.setFormatter(ApeColorFormatter())
    cli_logger.handlers = [handler]
    return cli_logger


logger = _get_logger("ape")


__all__ = ["logger"]



from typing import List, Optional, Type

from ape.types import AddressType

from ..exceptions import AddressError
from .base import abstractdataclass, abstractmethod
from .providers import ProviderAPI, ReceiptAPI, TransactionAPI


@abstractdataclass
class AddressAPI:
    _provider: Optional[ProviderAPI] = None

    @property
    def provider(self) -> ProviderAPI:
        if not self._provider:
            raise AddressError(
                f"Incorrectly implemented provider API for class {type(self).__name__}"
            )

        return self._provider

    @provider.setter
    def provider(self, value: ProviderAPI):
        self._provider = value

    @property
    def _receipt_class(self) -> Type[ReceiptAPI]:
        return self.provider.network.ecosystem.receipt_class

    @property
    def _transaction_class(self) -> Type[TransactionAPI]:
        return self.provider.network.ecosystem.transaction_class

    @property
    @abstractmethod
    def address(self) -> AddressType:
        ...

    def __dir__(self) -> List[str]:
        # This displays methods to IPython on `a.[TAB]` tab completion
        return [
            "address",
            "balance",
            "code",
            "codesize",
            "nonce",
            "is_contract",
            "provider",
        ]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.address}>"

    def __str__(self) -> str:
        return self.address

    @property
    def nonce(self) -> int:
        return self.provider.get_nonce(self.address)

    @property
    def balance(self) -> int:
        return self.provider.get_balance(self.address)

    @property
    def code(self) -> bytes:
        # TODO: Explore caching this (based on `self.provider.network` and examining code)
        return self.provider.get_code(self.address)

    @property
    def codesize(self) -> int:
        return len(self.code)

    @property
    def is_contract(self) -> bool:
        return len(self.code) > 0


class Address(AddressAPI):
    _address: AddressType

    @property
    def address(self) -> AddressType:
        return self._address



import difflib
import re
from typing import Any, Dict

import click
import yaml

from ape.cli import Abort, ape_cli_context
from ape.exceptions import ApeException
from ape.logging import logger
from ape.plugins import clean_plugin_name

try:
    from importlib import metadata  # type: ignore
except ImportError:
    import importlib_metadata as metadata  # type: ignore


_DIFFLIB_CUT_OFF = 0.6


def display_config(ctx, param, value):
    # NOTE: This is necessary not to interrupt how version or help is intercepted
    if not value or ctx.resilient_parsing:
        return

    from ape import project

    click.echo("# Current configuration")
    click.echo(yaml.dump(project.config.serialize()))

    ctx.exit()  # NOTE: Must exit to bypass running ApeCLI


class ApeCLI(click.MultiCommand):
    _commands = None

    def invoke(self, ctx) -> Any:
        try:
            return super().invoke(ctx)
        except click.UsageError as err:
            self._suggest_cmd(err)
        except ApeException as err:
            raise Abort(str(err)) from err

    @staticmethod
    def _suggest_cmd(usage_error):
        if usage_error.message is None:
            raise usage_error

        match = re.match("No such command '(.*)'.", usage_error.message)
        if not match:
            raise usage_error

        bad_arg = match.groups()[0]
        suggested_commands = difflib.get_close_matches(
            bad_arg, list(usage_error.ctx.command.commands.keys()), cutoff=_DIFFLIB_CUT_OFF
        )
        if suggested_commands:
            if bad_arg not in suggested_commands:
                usage_error.message = (
                    f"No such command '{bad_arg}'. Did you mean {' or '.join(suggested_commands)}?"
                )

        raise usage_error

    @property
    def commands(self) -> Dict:
        if not self._commands:
            entry_points = metadata.entry_points()  # type: ignore

            if "ape_cli_subcommands" not in entry_points:
                raise Abort("Missing registered cli subcommands")

            self._commands = {
                clean_plugin_name(entry_point.name): entry_point.load
                for entry_point in entry_points["ape_cli_subcommands"]
            }

        return self._commands

    def list_commands(self, ctx):
        return list(sorted(self.commands))

    def get_command(self, ctx, name):
        if name in self.commands:
            try:
                return self.commands[name]()
            except Exception as err:
                logger.warning(f"Unable to load CLI endpoint for plugin 'ape_{name}'.\n\t{err}")

        # NOTE: don't return anything so Click displays proper error


@click.command(cls=ApeCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@ape_cli_context()
@click.version_option(message="%(version)s", package_name="eth-ape")
@click.option(
    "--config",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_config,
    help="Show configuration options (using `ape-config.yaml`)",
)
def cli(context):
    _ = context



from eth_typing import ChecksumAddress as AddressType

from .contract import ABI, Bytecode, Checksum, Compiler, ContractType, Source
from .manifest import PackageManifest, PackageMeta
from .signatures import MessageSignature, SignableMessage, TransactionSignature

__all__ = [
    "ABI",
    "AddressType",
    "Bytecode",
    "Checksum",
    "Compiler",
    "ContractType",
    "MessageSignature",
    "PackageManifest",
    "PackageMeta",
    "SignableMessage",
    "Source",
    "TransactionSignature",
]



from ape import plugins
from ape.api.config import ConfigEnum, ConfigItem


class EvmVersion(ConfigEnum):
    constaninople = "constaninople"
    istanbul = "istanbul"


class Config(ConfigItem):
    evm_version: EvmVersion = EvmVersion.istanbul

    def validate_config(self):
        self.evm_version = EvmVersion[self.evm_version]


@plugins.register(plugins.Config)
def config_class():
    return Config



import functools
import importlib
import pkgutil
from typing import Any, Callable, Iterator, Tuple, Type, cast

from ape.logging import logger

from .account import AccountPlugin
from .compiler import CompilerPlugin
from .config import Config
from .converter import ConversionPlugin
from .network import EcosystemPlugin, ExplorerPlugin, NetworkPlugin, ProviderPlugin
from .pluggy_patch import PluginType, hookimpl, plugin_manager


class PluginError(Exception):
    pass


# Combine all the plugins together via subclassing (merges `hookspec`s)
class AllPluginHooks(
    Config,
    AccountPlugin,
    CompilerPlugin,
    ConversionPlugin,
    EcosystemPlugin,
    ExplorerPlugin,
    NetworkPlugin,
    ProviderPlugin,
):
    pass


# All hookspecs are registered
plugin_manager.add_hookspecs(AllPluginHooks)

# Add cast so that mypy knows that pm.hook is actually a `Plugins` instance.
# Without this hint there really is no way for mypy to know this.
plugin_manager.hook = cast(AllPluginHooks, plugin_manager.hook)


def clean_plugin_name(name: str) -> str:
    return name.replace("ape_", "").replace("_", "-")


def get_hooks(plugin_type):
    return [name for name, method in plugin_type.__dict__.items() if hasattr(method, "ape_spec")]


def register(plugin_type: Type[PluginType], **hookimpl_kwargs) -> Callable:
    # NOTE: we are basically checking that `plugin_type`
    #       is one of the parent classes of `Plugins`
    if not issubclass(AllPluginHooks, plugin_type):
        raise PluginError("Not a valid plugin type to register")

    def check_hook(plugin_type, hookimpl_kwargs, fn):
        fn = hookimpl(fn, **hookimpl_kwargs)

        if not hasattr(plugin_type, fn.__name__):
            hooks = get_hooks(plugin_type)

            raise PluginError(
                f"Registered function `{fn.__name__}` is not"
                f" a valid hook for {plugin_type.__name__}, must be one of:"
                f" {hooks}"
            )

        return fn

    # NOTE: Get around issue with using `plugin_type` raw in `check_hook`
    return functools.partial(check_hook, plugin_type, hookimpl_kwargs)


def valid_impl(api_class: Any) -> bool:
    if isinstance(api_class, tuple):
        return all(valid_impl(c) for c in api_class)

    # Is not an ABC base class or abstractdataclass
    if not hasattr(api_class, "__abstractmethods__"):
        return True  # not an abstract class

    else:
        return len(api_class.__abstractmethods__) == 0


class PluginManager:
    def __init__(self):
        # NOTE: This actually loads the plugins, and should only be done once
        for _, name, ispkg in pkgutil.iter_modules():
            if name.startswith("ape_") and ispkg:
                try:
                    plugin_manager.register(importlib.import_module(name))
                except Exception:
                    logger.warning(f"Error loading plugin package '{name}'")

    def __getattr__(self, attr_name: str) -> Iterator[Tuple[str, tuple]]:
        if not hasattr(plugin_manager.hook, attr_name):
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")

        # Do this to get access to the package name
        hook_fn = getattr(plugin_manager.hook, attr_name)
        hookimpls = hook_fn.get_hookimpls()

        def get_plugin_name_and_hookfn(h):
            return h.plugin_name, getattr(h.plugin, attr_name)()

        for plugin_name, results in map(get_plugin_name_and_hookfn, hookimpls):
            # NOTE: Some plugins return a tuple and some return iterators
            if not isinstance(results, tuple) and hasattr(results, "__iter__"):
                # Only if it's an iterator, provider results as a series
                for result in results:
                    if valid_impl(result):
                        yield clean_plugin_name(plugin_name), result

                    else:
                        logger.warning(
                            f"'{result.__name__}' from '{plugin_name}' is not fully implemented"
                        )

            elif valid_impl(results):
                # Otherwise, provide results directly
                yield clean_plugin_name(plugin_name), results

            else:
                _warn_not_fully_implemented_error(results, plugin_name)


def _warn_not_fully_implemented_error(results, plugin_name):
    if isinstance(results, (tuple, list)):
        class_names = ", ".join((r.__name__ for r in results))
    else:
        class_names = results.__name__
    logger.warning(f"'{class_names}' from '{plugin_name}' is not fully implemented")


__all__ = [
    "PluginManager",
    "clean_plugin_name",
    "register",
]



import pytest  # type: ignore
import requests
from hypothesis import HealthCheck, given, settings
from hypothesis_jsonschema import from_schema  # type: ignore

from ape.types.manifest import PackageManifest

ETHPM_MANIFEST_SCHEMA_URI = (
    "https://raw.githubusercontent.com/ethpm/ethpm-spec/master/spec/v3.spec.json"
)


@pytest.mark.xfail(reason="Schema is poorly formed")
@pytest.mark.fuzzing
@given(manifest_dict=from_schema(requests.get(ETHPM_MANIFEST_SCHEMA_URI).json()))
@settings(suppress_health_check=(HealthCheck.too_slow,))
def test_manifest_parsing(manifest_dict):
    manifest = PackageManifest.from_dict(manifest_dict)
    assert manifest.to_dict() == manifest_dict


def test_example_manifests(manifest):
    assert PackageManifest.from_dict(manifest).to_dict() == manifest



import sys
from importlib import import_module
from pathlib import Path

import click

from ape import config, networks
from ape.cli import ape_cli_context, network_option, verbose_option
from ape.utils import get_relative_path
from ape_console._cli import console

# TODO: Migrate this to a CLI toolkit under ``ape``


def _run_script(cli_ctx, script_path, interactive=False, verbose=False):
    script_path = get_relative_path(script_path, Path.cwd())
    script_parts = script_path.parts[:-1]

    if any(p == ".." for p in script_parts):
        cli_ctx.abort("Cannot execute script from outside current directory")

    # Add to Python path so we can search for the given script to import
    sys.path.append(str(script_path.parent.resolve()))

    # Load the python module to find our hook functions
    try:
        py_module = import_module(script_path.stem)

    except Exception as err:
        if verbose:
            raise err

        else:
            cli_ctx.abort(f"Exception while executing script: {script_path}", base_error=err)

    finally:
        # Undo adding the path to make sure it's not permanent
        sys.path.remove(str(script_path.parent.resolve()))

    # Execute the hooks
    if hasattr(py_module, "cli"):
        # TODO: Pass context to ``cli`` before calling it
        py_module.cli()

    elif hasattr(py_module, "main"):
        # NOTE: ``main()`` accepts no arguments
        py_module.main()

    else:
        cli_ctx.abort("No `main` or `cli` method detected")

    if interactive:
        return console()


@click.command(short_help="Run scripts from the `scripts` folder")
@click.argument("scripts", nargs=-1)
@verbose_option(help="Display errors from scripts")
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    default=False,
    help="Drop into interactive console session after running",
)
@network_option
@ape_cli_context()
def cli(cli_ctx, scripts, verbose, interactive, network):
    """
    NAME - Path or script name (from ``scripts/`` folder)

    Run scripts from the ``scripts`` folder. A script must either define a ``main()`` method,
    or define an import named ``cli`` that is a ``click.Command`` or ``click.Group`` object.
    ``click.Group`` and ``click.Command`` objects will be provided with additional context, which
    will be injected dynamically during script execution. The dynamically injected objects are
    the exports from the ``ape`` top-level package (similar to how the console works)
    """
    if not scripts:
        cli_ctx.abort("Must provide at least one script name or path")

    scripts_folder = config.PROJECT_FOLDER / "scripts"

    # Generate the lookup based on all the scripts defined in the project's ``scripts/`` folder
    # NOTE: If folder does not exist, this will be empty (same as if there are no files)
    available_scripts = {p.stem: p.resolve() for p in scripts_folder.glob("*.py")}

    with networks.parse_network_choice(network):
        for name in scripts:
            if Path(name).exists():
                script_file = Path(name).resolve()

            elif not scripts_folder.exists():
                cli_ctx.abort("No `scripts/` directory detected to run script")

            elif name not in available_scripts:
                cli_ctx.abort(f"No script named '{name}' detected in scripts folder")

            else:
                script_file = available_scripts[name]

            _run_script(cli_ctx, script_file, interactive, verbose)



_DEFAULT_NETWORKS_LIST = """
ecosystems:
- name: ethereum  # Default
  - name: mainnet
    providers:
    - http  # Default
  - name: ropsten
    providers:
    - http  # Default
  - name: kovan
    providers:
    - http  # Default
  - name: rinkeby
    providers:
    - http  # Default
  - name: goerli
    providers:
    - http  # Default
  - name: development  # Default
    providers:
    - http  # Default
    - test
"""


def test_list(ape_cli, runner):
    result = runner.invoke(ape_cli, ["networks", "list"])
    expected = _DEFAULT_NETWORKS_LIST.strip()
    assert expected in result.output



from typing import Dict, Iterator, Optional

from dataclassy import dataclass
from pluggy import PluginManager  # type: ignore

from ape.api import EcosystemAPI, ProviderAPI, ProviderContextManager
from ape.exceptions import NetworkError
from ape.utils import cached_property

from .config import ConfigManager


@dataclass
class NetworkManager:
    """
    The set of all blockchain network ecosystems that Ape has registered through its plugin system.
    """

    config: ConfigManager
    plugin_manager: PluginManager
    active_provider: Optional[ProviderAPI] = None
    _default: Optional[str] = None

    @cached_property
    def ecosystems(self) -> Dict[str, EcosystemAPI]:
        return {
            plugin_name: ecosystem_class(
                name=plugin_name,
                network_manager=self,
                config_manager=self.config,
                plugin_manager=self.plugin_manager,
                data_folder=self.config.DATA_FOLDER / plugin_name,
                request_header=self.config.REQUEST_HEADER,
            )
            for plugin_name, ecosystem_class in self.plugin_manager.ecosystems
        }

    def __iter__(self) -> Iterator[str]:
        yield from self.ecosystems

    def __getitem__(self, ecosystem_name: str) -> EcosystemAPI:
        if ecosystem_name not in self.ecosystems:
            raise NetworkError(f"Unknown ecosystem '{ecosystem_name}'")

        return self.ecosystems[ecosystem_name]

    def __getattr__(self, attr_name: str) -> EcosystemAPI:
        if attr_name not in self.ecosystems:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")

        return self.ecosystems[attr_name]

    @property
    def network_choices(self) -> Iterator[str]:
        """
        Produce the set of all possible network choices that could be provided
        for a "network selection" choice e.g. `--network [ECOSYSTEM:NETWORK:PROVIDER]`
        """
        for ecosystem_name, ecosystem in self.ecosystems.items():
            yield ecosystem_name
            for network_name, network in ecosystem.networks.items():
                if ecosystem_name == self.default_ecosystem.name:
                    yield f":{network_name}"

                yield f"{ecosystem_name}:{network_name}"

                for provider in network.providers:
                    if (
                        ecosystem_name == self.default_ecosystem.name
                        and network_name == ecosystem.default_network
                    ):
                        yield f"::{provider}"

                    elif ecosystem_name == self.default_ecosystem.name:
                        yield f":{network_name}:{provider}"

                    elif network_name == ecosystem.default_network:
                        yield f"{ecosystem_name}::{provider}"

                    yield f"{ecosystem_name}:{network_name}:{provider}"

    def parse_network_choice(
        self,
        network_choice: Optional[str] = None,
    ) -> ProviderContextManager:
        if network_choice is None:
            return self.default["development"].use_default_provider()

        selections = network_choice.split(":")

        # NOTE: Handle case when URI is passed e.g. "http://..."
        if len(selections) > 3:
            selections[2] = ":".join(selections[2:])

        if selections == network_choice or len(selections) == 1:
            # Either split didn't work (in which case it matches the start)
            # or there was nothing after the ``:`` (e.g. "ethereum:")
            ecosystem = self.__getattr__(selections[0] or self.default_ecosystem.name)
            # By default, the "development" network should be specified for
            # any ecosystem (this should not correspond to a production chain)
            return ecosystem["development"].use_default_provider()

        elif len(selections) == 2:
            # Only ecosystem and network were specified, not provider
            ecosystem_name, network_name = selections
            ecosystem = self.__getattr__(ecosystem_name or self.default_ecosystem.name)
            network = ecosystem[network_name or ecosystem.default_network]
            return network.use_default_provider()

        elif len(selections) == 3:
            # Everything is specified, use specified provider for ecosystem
            # and network
            ecosystem_name, network_name, provider_name = selections
            ecosystem = self.__getattr__(ecosystem_name or self.default_ecosystem.name)
            network = ecosystem[network_name or ecosystem.default_network]
            return network.use_provider(provider_name)

        else:
            # NOTE: Might be unreachable
            raise NetworkError("Invalid network selection")

    @property
    def default_ecosystem(self) -> EcosystemAPI:
        if self._default:
            return self.ecosystems[self._default]

        # If explicit default is not set, use first registered ecosystem
        elif len(self.ecosystems) == 1:
            return self.ecosystems[list(self.__iter__())[0]]

        else:
            raise NetworkError("No ecosystems installed")

    def set_default_ecosystem(self, ecosystem_name: str):
        if ecosystem_name in self.__iter__():
            self._default = ecosystem_name

        else:
            raise NetworkError("Not a registered ecosystem")



import json
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import dataclassy as dc


def update_params(params, param_name, param_type):
    if param_name in params and params[param_name]:
        params[param_name] = param_type.from_dict(params[param_name])


def update_list_params(params, param_name, param_type):
    if param_name in params and params[param_name]:
        params[param_name] = [param_type.from_dict(p) for p in params[param_name]]


def update_dict_params(params, param_name, param_type):
    if param_name in params and params[param_name]:
        for key in params[param_name]:
            params[param_name][key] = param_type.from_dict(params[param_name][key])


def remove_empty_fields(data, keep_fields: Optional[Set[str]] = None):
    if isinstance(data, dict):
        return {
            k: v
            for k, v in zip(data.keys(), map(remove_empty_fields, data.values()))
            if isinstance(v, bool) or (keep_fields and k in keep_fields) or v
        }

    elif isinstance(data, list):
        return [v for v in map(remove_empty_fields, data) if isinstance(v, bool) or v]

    return data


def to_dict(v: Any) -> Optional[Union[list, dict, str, int, bool]]:
    if isinstance(v, SerializableType):
        return v.to_dict()

    elif isinstance(v, list):
        return [to_dict(i) for i in v]  # type: ignore

    elif isinstance(v, dict):
        return {k: to_dict(i) for k, i in v.items()}

    elif isinstance(v, (str, int, bool)) or v is None:
        return v

    else:
        raise ValueError(f"Unhandled type '{type(v)}'")


@dc.dataclass(slots=True, kwargs=True, repr=True)
class SerializableType:
    _keep_fields_: Set[str] = set()
    _skip_fields_: Set[str] = set()

    def to_dict(self) -> Dict:
        data = {
            k: to_dict(v)
            for k, v in dc.values(self).items()
            if not (k.startswith("_") or k in self._skip_fields_)
        }
        return remove_empty_fields(data, keep_fields=self._keep_fields_)

    @classmethod
    def from_dict(cls, params: Dict):
        params = deepcopy(params)
        return cls(**params)  # type: ignore


class FileMixin(SerializableType):
    @classmethod
    def from_file(cls, path: Path):
        with path.open("r") as f:
            return cls.from_dict(json.load(f))

    def to_file(self, path: Path):
        # NOTE: EIP-2678 specifies document *must* be tightly packed
        # NOTE: EIP-2678 specifies document *must* have sorted keys
        with path.open("w") as f:
            json.dump(self.to_dict(), f, indent=4, sort_keys=True)



import os
from distutils.dir_util import copy_tree
from pathlib import Path

import pytest  # type: ignore
from click.testing import CliRunner

import ape
from ape import Project

TEST_PROJECTS_FOLDER = Path(__file__).parent / "data" / "projects"


@pytest.fixture(params=[p for p in TEST_PROJECTS_FOLDER.iterdir() if p.is_dir()])
def project_folder(request, config):
    project_source_dir = request.param
    project_dest_dir = config.PROJECT_FOLDER / project_source_dir.name
    copy_tree(project_source_dir.as_posix(), project_dest_dir.as_posix())
    previous_project_folder = config.PROJECT_FOLDER
    config.PROJECT_FOLDER = project_dest_dir
    yield project_dest_dir
    config.PROJECT_FOLDER = previous_project_folder


@pytest.fixture
def project(project_folder):
    previous_project = ape.project
    project = Project(project_folder)
    ape.project = project
    yield project
    ape.project = previous_project


@pytest.fixture
def runner(project_folder):
    previous_cwd = str(Path.cwd())
    os.chdir(str(project_folder))
    runner = CliRunner()
    yield runner
    os.chdir(previous_cwd)


@pytest.fixture(scope="session")
def ape_cli():
    from ape._cli import cli

    yield cli


def assert_failure(result, expected_output):
    assert result.exit_code == 1
    assert result.exception is not None
    assert "ERROR" in result.output
    assert expected_output in result.output



from enum import IntEnum
from pathlib import Path
from typing import Iterator, List, Optional

from dataclassy import as_dict

from ape.logging import logger
from ape.types import TransactionSignature

from ..exceptions import ProviderError
from . import networks
from .base import abstractdataclass, abstractmethod
from .config import ConfigItem


@abstractdataclass
class TransactionAPI:
    chain_id: int = 0
    sender: str = ""
    receiver: str = ""
    nonce: Optional[int] = None  # NOTE: `Optional` only to denote using default behavior
    value: int = 0
    gas_limit: Optional[int] = None  # NOTE: `Optional` only to denote using default behavior
    gas_price: Optional[int] = None  # NOTE: `Optional` only to denote using default behavior
    data: bytes = b""

    signature: Optional[TransactionSignature] = None

    def __post_init__(self):
        if not self.is_valid:
            raise ProviderError("Transaction is not valid")

    @property
    @abstractmethod
    def is_valid(self):
        ...

    @abstractmethod
    def encode(self) -> bytes:
        """
        Take this object and produce a hash to sign to submit a transaction
        """

    def as_dict(self) -> dict:
        return as_dict(self)

    def __repr__(self) -> str:
        data = as_dict(self)  # NOTE: `as_dict` could be overriden
        params = ", ".join(f"{k}={v}" for k, v in data.items())
        return f"<{self.__class__.__name__} {params}>"

    def __str__(self) -> str:
        data = as_dict(self)  # NOTE: `as_dict` could be overriden
        if len(data["data"]) > 9:
            data["data"] = (
                "0x" + bytes(data["data"][:3]).hex() + "..." + bytes(data["data"][-3:]).hex()
            )
        else:
            data["data"] = "0x" + bytes(data["data"]).hex()
        params = "\n  ".join(f"{k}: {v}" for k, v in data.items())
        return f"{self.__class__.__name__}:\n  {params}"


class TransactionStatusEnum(IntEnum):
    failing = 0
    no_error = 1


@abstractdataclass
class ReceiptAPI:
    txn_hash: str
    status: TransactionStatusEnum
    block_number: int
    gas_used: int
    gas_price: int
    logs: List[dict] = []
    contract_address: Optional[str] = None

    def __post_init__(self):
        logger.info(f"Submitted {self.txn_hash.hex()}")

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.txn_hash}>"

    @classmethod
    @abstractmethod
    def decode(cls, data: dict) -> "ReceiptAPI":
        ...


@abstractdataclass
class ProviderAPI:
    """
    A Provider must work with a particular Network in a particular Ecosystem
    """

    name: str  # Plugin name
    network: networks.NetworkAPI
    config: ConfigItem
    provider_settings: dict
    data_folder: Path
    request_header: str

    @abstractmethod
    def connect(self):
        ...

    @abstractmethod
    def disconnect(self):
        ...

    @abstractmethod
    def update_settings(self, new_settings: dict):
        ...

    @property
    @abstractmethod
    def chain_id(self) -> int:
        ...

    @abstractmethod
    def get_balance(self, address: str) -> int:
        ...

    @abstractmethod
    def get_code(self, address: str) -> bytes:
        ...

    @abstractmethod
    def get_nonce(self, address: str) -> int:
        ...

    @abstractmethod
    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        ...

    @property
    @abstractmethod
    def gas_price(self) -> int:
        ...

    @abstractmethod
    def send_call(self, txn: TransactionAPI) -> bytes:  # Return value of function
        ...

    @abstractmethod
    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        ...

    @abstractmethod
    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        ...

    @abstractmethod
    def get_events(self, **filter_params) -> Iterator[dict]:
        ...



import faulthandler

import click
import IPython  # type: ignore

from ape import networks
from ape import project as default_project
from ape.cli import network_option, verbose_option
from ape.version import version as ape_version  # type: ignore


@click.command(short_help="Load the console", context_settings=dict(ignore_unknown_options=True))
@verbose_option(help="Display more information in the console")
@network_option
def cli(verbose, network):
    """
    Opens a console for the local project."""

    with networks.parse_network_choice(network):
        return console(verbose=verbose)


def console(project=None, verbose=False, extra_locals=None):
    import ape

    if not project:
        # Use default project
        project = default_project

    banner = ""

    if verbose:
        banner = """
   Python:  {python_version}
  IPython:  {ipython_version}
      Ape:  {ape_version}
  Project:  {project_path}

    Are you ready to Ape, anon?
    """.format(
            python_version=ape._python_version,
            ipython_version=IPython.__version__,
            ape_version=ape_version,
            project_path=project.path,
        )

        faulthandler.enable()  # NOTE: In case we segfault

    namespace = {component: getattr(ape, component) for component in ape.__all__}

    if extra_locals:
        namespace.update(extra_locals)

    IPython.embed(colors="Neutral", banner1=banner, user_ns=namespace)



from typing import TYPE_CHECKING, Any, Generic, TypeVar

from .base import abstractdataclass, abstractmethod
from .config import ConfigItem

if TYPE_CHECKING:
    from ape.managers.networks import NetworkManager

ConvertedType = TypeVar("ConvertedType")


@abstractdataclass
class ConverterAPI(Generic[ConvertedType]):
    # NOTE: In case we need to store info e.g. tokenlists
    config: ConfigItem

    # NOTE: In case we need access to a network e.g. ENS
    networks: "NetworkManager"

    @abstractmethod
    def is_convertible(self, value: Any) -> bool:
        """
        Returns `True` if string value provided by `value` is convertible using
        `self.convert(value)`
        """

    @abstractmethod
    def convert(self, value: Any) -> ConvertedType:
        """
        Implements any conversion logic on `value` to produce `ABIType`.

        Must throw if not convertible.
        """



from abc import ABCMeta, abstractmethod
from functools import partial

from dataclassy import dataclass
from dataclassy.dataclass import DataClassMeta


class AbstractDataClassMeta(DataClassMeta, ABCMeta):
    pass


abstractdataclass = partial(dataclass, kwargs=True, meta=AbstractDataClassMeta)


__all__ = [
    "abstractdataclass",
    "abstractmethod",
    "AbstractDataClassMeta",
    "dataclass",
]



import json

import click
from eth_account import Account as EthAccount  # type: ignore
from eth_utils import to_bytes

from ape import accounts
from ape.cli import ape_cli_context, existing_alias_argument, non_existing_alias_argument
from ape_accounts import KeyfileAccount

# NOTE: Must used the instantiated version of `AccountsContainer` in `accounts`
container = accounts.containers["accounts"]


@click.group(short_help="Manage local accounts")
def cli():
    """
    Command-line helper for managing local accounts. You can unlock local accounts from
    scripts or the console using the accounts.load() method.
    """


# Different name because `list` is a keyword
@cli.command(name="list", short_help="List available local accounts")
@click.option("--all", help="Output accounts from all plugins", is_flag=True)
@ape_cli_context()
def _list(cli_ctx, all):
    accounts_to_output = accounts if all else accounts.containers.get("accounts", [])
    if len(accounts_to_output) == 0:
        cli_ctx.logger.warning("No accounts found.")
        return

    elif len(accounts_to_output) > 1:
        click.echo(f"Found {len(accounts)} accounts:")

    else:
        click.echo("Found 1 account:")

    for account in accounts_to_output:
        alias_display = f" (alias: '{account.alias}')" if account.alias else ""
        click.echo(f"  {account.address}{alias_display}")


@cli.command(short_help="Create a new keyfile account with a random private key")
@non_existing_alias_argument
@ape_cli_context()
def generate(cli_ctx, alias):
    path = container.data_folder.joinpath(f"{alias}.json")
    extra_entropy = click.prompt(
        "Add extra entropy for key generation...",
        hide_input=True,
    )
    account = EthAccount.create(extra_entropy)
    passphrase = click.prompt(
        "Create Passphrase",
        hide_input=True,
        confirmation_prompt=True,
    )
    path.write_text(json.dumps(EthAccount.encrypt(account.key, passphrase)))
    cli_ctx.logger.success(
        f"A new account '{account.address}' has been added with the id '{alias}'"
    )


# Different name because `import` is a keyword
@cli.command(name="import", short_help="Add a new keyfile account by entering a private key")
@non_existing_alias_argument
@ape_cli_context()
def _import(cli_ctx, alias):
    path = container.data_folder.joinpath(f"{alias}.json")
    key = click.prompt("Enter Private Key", hide_input=True)
    try:
        account = EthAccount.from_key(to_bytes(hexstr=key))
    except Exception as error:
        cli_ctx.abort(f"Key can't be imported: {error}")
        return

    passphrase = click.prompt(
        "Create Passphrase",
        hide_input=True,
        confirmation_prompt=True,
    )
    path.write_text(json.dumps(EthAccount.encrypt(account.key, passphrase)))
    cli_ctx.logger.success(
        f"A new account '{account.address}' has been added with the id '{alias}'"
    )


@cli.command(short_help="Change the password of an existing account")
@existing_alias_argument(account_type=KeyfileAccount)
@ape_cli_context()
def change_password(cli_ctx, alias):
    account = accounts.load(alias)
    account.change_password()
    cli_ctx.logger.success(f"Password has been changed for account '{alias}'")


@cli.command(short_help="Delete an existing account")
@existing_alias_argument(account_type=KeyfileAccount)
@ape_cli_context()
def delete(cli_ctx, alias):
    account = accounts.load(alias)
    account.delete()
    cli_ctx.logger.success(f"Account '{alias}' has been deleted")



from typing import Tuple, Type

from ape.api.compiler import CompilerAPI

from .pluggy_patch import PluginType, hookspec


class CompilerPlugin(PluginType):
    @hookspec
    def register_compiler(self) -> Tuple[Tuple[str], Type[CompilerAPI]]:
        """
        Returns a set of file extensions the plugin handles,
        and the compiler class that can be used to compile them.
        """



import signal

signal.signal(signal.SIGINT, lambda s, f: _sys.exit(130))

import sys as _sys
from functools import partial as _partial
from pathlib import Path as _Path

from .api.contracts import _Contract
from .managers.accounts import AccountManager as _AccountManager
from .managers.compilers import CompilerManager as _CompilerManager
from .managers.config import ConfigManager as _ConfigManager
from .managers.converters import ConversionManager as _ConversionManager
from .managers.networks import NetworkManager as _NetworkManager
from .managers.project import ProjectManager as _ProjectManager
from .plugins import PluginManager as _PluginManager
from .utils import get_package_version

__version__ = get_package_version(__name__)


# NOTE: DO NOT OVERWRITE
_python_version = (
    f"{_sys.version_info.major}.{_sys.version_info.minor}"
    f".{_sys.version_info.micro} {_sys.version_info.releaselevel}"
)

# Wiring together the application

plugin_manager = _PluginManager()
"""Manages plugins for the current project. See :class:`ape.plugins.PluginManager`."""

config = _ConfigManager(  # type: ignore
    # Store all globally-cached files
    DATA_FOLDER=_Path.home().joinpath(".ape"),
    # NOTE: For all HTTP requests we make
    REQUEST_HEADER={
        "User-Agent": f"Ape/{__version__} (Python/{_python_version})",
    },
    # What we are considering to be the starting project directory
    PROJECT_FOLDER=_Path.cwd(),
    plugin_manager=plugin_manager,
)
"""The active configs for the current project. See :class:`ape.managers.config.ConfigManager`."""

# Main types we export for the user

compilers = _CompilerManager(config, plugin_manager)  # type: ignore
"""Manages compilers for the current project. See
:class:`ape.managers.compilers.CompilerManager`."""

networks = _NetworkManager(config, plugin_manager)  # type: ignore
"""Manages the networks for the current project. See
:class:`ape.managers.networks.NetworkManager`."""

_converters = _ConversionManager(config, plugin_manager, networks)  # type: ignore

accounts = _AccountManager(config, _converters, plugin_manager, networks)  # type: ignore
"""Manages accounts for the current project. See :class:`ape.managers.accounts.AccountManager`."""

Project = _partial(_ProjectManager, config=config, compilers=compilers)
"""User-facing class for instantiating Projects (in addition to the currently
active ``project``). See :class:`ape.managers.project.ProjectManager`."""

project = Project(config.PROJECT_FOLDER)
"""The currently active project. See :class:`ape.managers.project.ProjectManager`."""

Contract = _partial(_Contract, networks=networks, converters=_converters)
"""User-facing class for instantiating contracts. See :class:`ape.api.contracts._Contract`."""

convert = _converters.convert
"""Conversion utility function. See :class:`ape.managers.converters.ConversionManager`."""

__all__ = [
    "accounts",
    "compilers",
    "config",
    "convert",
    "Contract",
    "networks",
    "project",
    "Project",  # So you can load other projects
]



from copy import deepcopy
from typing import Dict, List, Optional

from .abstract import (
    FileMixin,
    SerializableType,
    update_dict_params,
    update_list_params,
    update_params,
)
from .contract import Compiler, ContractInstance, ContractType, Source


class PackageMeta(SerializableType):
    authors: Optional[List[str]] = None
    license: Optional[str] = None
    description: Optional[str] = None
    keywords: Optional[List[str]] = None
    links: Optional[Dict[str, str]] = None


class PackageManifest(FileMixin, SerializableType):
    # NOTE: Must not override this key
    manifest: str = "ethpm/3"
    # NOTE: ``name`` and ``version`` should appear together
    # NOTE: ``name`` must begin lowercase, and be comprised of only ``[a-z0-9-]`` chars
    # NOTE: ``name`` should not exceed 255 chars in length
    name: Optional[str] = None
    # NOTE: ``version`` should be valid SemVer
    version: Optional[str] = None
    # NOTE: ``meta`` should be in all published packages
    meta: Optional[PackageMeta] = None
    # NOTE: ``sources`` source tree should be necessary and sufficient to compile
    #       all ``ContractType``s in manifest
    sources: Optional[Dict[str, Source]] = None
    # NOTE: ``contractTypes`` should only include types directly computed from manifest
    # NOTE: ``contractTypes`` should not include abstracts
    contractTypes: Optional[Dict[str, ContractType]] = None
    compilers: Optional[List[Compiler]] = None
    # NOTE: Keys must be a valid BIP122 URI chain definition
    # NOTE: Values must be a dict of ``ContractType.contractName`` => ``ContractInstance`` objects
    deployments: Optional[Dict[str, Dict[str, ContractInstance]]] = None
    # NOTE: keys must begin lowercase, and be comprised of only ``[a-z0-9-]`` chars
    #       (like ``PackageManifest.name``)
    # NOTE: keys should not exceed 255 chars in length (like ``PackageManifest.name``)
    # NOTE: values must be a Content Addressible URI that conforms to the same manifest
    #       version as ``manifest``
    buildDependencies: Optional[Dict[str, str]] = None

    def __getattr__(self, attr_name: str):
        if self.contractTypes and attr_name in self.contractTypes:
            return self.contractTypes[attr_name]

        else:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")

    @classmethod
    def from_dict(cls, params: Dict):
        params = deepcopy(params)
        update_params(params, "meta", PackageMeta)
        update_dict_params(params, "sources", Source)

        # NOTE: Special 1-level dict with key in type as arg
        if "contractTypes" in params and params["contractTypes"]:
            for name in params["contractTypes"]:
                params["contractTypes"][name] = ContractType.from_dict(  # type: ignore
                    {
                        # NOTE: We inject this parameter ourselves, remove it when serializing
                        "contractName": name,
                        **params["contractTypes"][name],
                    }
                )

        update_list_params(params, "compilers", Compiler)

        # NOTE: Special 2-level dict
        if "deployments" in params and params["deployments"]:
            for name in params["deployments"]:
                update_dict_params(params["deployments"], name, ContractInstance)

        return cls(**params)  # type: ignore



import json
from pathlib import Path

import pytest  # type: ignore
from eth_account import Account  # type: ignore

from ape.api import AccountAPI

from .conftest import assert_failure

ALIAS = "test"
PASSWORD = "a"
PRIVATE_KEY = "0000000000000000000000000000000000000000000000000000000000000001"
IMPORT_VALID_INPUT = "\n".join([f"0x{PRIVATE_KEY}", PASSWORD, PASSWORD])
GENERATE_VALID_INPUT = "\n".join(["random entropy", PASSWORD, PASSWORD])
MOCK_LOCAL_ALIAS = "test_local_alias"
MOCK_LOCAL_ADDRESS = "test_local_address"
MOCK_EXTERNAL_ALIAS = "test_external_alias"
MOCK_EXTERNAL_ADDRESS = "test_external_address"


@pytest.fixture
def test_keyparams():
    # NOTE: password is 'a'
    return {
        "address": "7e5f4552091a69125d5dfcb7b8c2659029395bdf",
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {"iv": "7bc492fb5dca4fe80fd47645b2aad0ff"},
            "ciphertext": "43beb65018a35c31494f642ec535315897634b021d7ec5bb8e0e2172387e2812",
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 262144,
                "r": 1,
                "p": 8,
                "salt": "4b127cb5ddbc0b3bd0cc0d2ef9a89bec",
            },
            "mac": "6a1d520975a031e11fc16cff610f5ae7476bcae4f2f598bc59ccffeae33b1caa",
        },
        "id": "ee424db9-da20-405d-bd75-e609d3e2b4ad",
        "version": 3,
    }


@pytest.fixture(autouse=True)
def test_keyfile_path(config):
    test_keyfile_path = Path(config.DATA_FOLDER / "accounts" / f"{ALIAS}.json")

    if test_keyfile_path.exists():
        # Corrupted from a previous test
        test_keyfile_path.unlink()

    return test_keyfile_path


@pytest.fixture
def test_keyfile(test_keyfile_path, test_keyparams):
    test_keyfile_path.write_text(json.dumps(test_keyparams))

    yield test_keyfile_path

    if test_keyfile_path.exists():
        test_keyfile_path.unlink()


@pytest.fixture
def test_account():
    return Account.from_key(bytes.fromhex(PRIVATE_KEY))


@pytest.fixture
def mock_local_account(mocker):
    mock_account = mocker.MagicMock(spec=AccountAPI)
    mock_account.alias = MOCK_LOCAL_ALIAS
    mock_account.address = MOCK_LOCAL_ADDRESS
    return mock_account


@pytest.fixture
def mock_third_party_account(mocker):
    mock_account = mocker.MagicMock(spec=AccountAPI)
    mock_account.alias = MOCK_EXTERNAL_ALIAS
    mock_account.address = MOCK_EXTERNAL_ADDRESS
    return mock_account


@pytest.fixture
def mock_account_manager(mocker, mock_local_account, mock_third_party_account):
    mock = mocker.patch("ape_accounts._cli.accounts")
    containers = {"accounts": [mock_local_account], "test-wallet": [mock_third_party_account]}
    mock.containers = containers
    return mock


def test_import(ape_cli, runner, test_account, test_keyfile_path):
    assert not test_keyfile_path.exists()
    # Add account from private keys
    result = runner.invoke(ape_cli, ["accounts", "import", ALIAS], input=IMPORT_VALID_INPUT)
    assert result.exit_code == 0, result.output
    assert test_account.address in result.output
    assert ALIAS in result.output
    assert test_keyfile_path.exists()


def test_import_alias_already_in_use(ape_cli, runner, test_account, test_keyfile_path):
    def invoke_import():
        return runner.invoke(ape_cli, ["accounts", "import", ALIAS], input=IMPORT_VALID_INPUT)

    result = invoke_import()
    assert result.exit_code == 0, result.output
    result = invoke_import()
    assert_failure(result, f"Account with alias '{ALIAS}' already in use")


def test_import_account_instantiation_failure(
    mocker, ape_cli, runner, test_account, test_keyfile_path
):
    eth_account_from_key_patch = mocker.patch("ape_accounts._cli.EthAccount.from_key")
    eth_account_from_key_patch.side_effect = Exception("Can't instantiate this account!")
    result = runner.invoke(ape_cli, ["accounts", "import", ALIAS], input=IMPORT_VALID_INPUT)
    assert_failure(result, "Key can't be imported: Can't instantiate this account!")


def test_generate(ape_cli, runner, test_keyfile_path):
    assert not test_keyfile_path.exists()
    # Generate new private key
    result = runner.invoke(ape_cli, ["accounts", "generate", ALIAS], input=GENERATE_VALID_INPUT)
    assert result.exit_code == 0, result.output
    assert ALIAS in result.output
    assert test_keyfile_path.exists()


def test_generate_alias_already_in_use(ape_cli, runner, test_account, test_keyfile_path):
    def invoke_generate():
        return runner.invoke(ape_cli, ["accounts", "generate", ALIAS], input=GENERATE_VALID_INPUT)

    result = invoke_generate()
    assert result.exit_code == 0, result.output
    result = invoke_generate()
    assert_failure(result, f"Account with alias '{ALIAS}' already in use")


def test_list(ape_cli, runner, test_keyfile):
    # Check availability
    assert test_keyfile.exists()
    result = runner.invoke(ape_cli, ["accounts", "list"])
    assert ALIAS in result.output


def test_list_excludes_external_accounts(ape_cli, runner, mock_account_manager):
    result = runner.invoke(ape_cli, ["accounts", "list"])
    assert result.exit_code == 0, result.output
    assert "test_local_alias" in result.output
    assert "test_local_address" in result.output
    assert "test_external_alias" not in result.output
    assert "test_external_address" not in result.output


def test_change_password(ape_cli, runner, test_keyfile):
    assert test_keyfile.exists()
    # Delete Account (`N` for "Leave unlocked?")
    valid_input = [PASSWORD, "N", "b", "b"]
    result = runner.invoke(
        ape_cli, ["accounts", "change-password", ALIAS], input="\n".join(valid_input) + "\n"
    )
    assert result.exit_code == 0, result.output


def test_delete(ape_cli, runner, test_keyfile):
    assert test_keyfile.exists()
    # Delete Account
    result = runner.invoke(ape_cli, ["accounts", "delete", ALIAS], input=PASSWORD + "\n")
    assert result.exit_code == 0, result.output
    assert not test_keyfile.exists()



import json
from pathlib import Path

import pytest  # type: ignore


@pytest.fixture(
    scope="session",
    params=[
        # Copied from https://github.com/ethpm/ethpm-spec/tree/master/examples
        "escrow",
        "owned",
        "piper-coin",
        "safe-math-lib",
        "standard-token",
        "transferable",
        "wallet-with-send",
        "wallet",
    ],
)
def manifest(request):
    yield json.loads((Path(__file__).parent / "data" / f"{request.param}.json").read_text())



from typing import Optional, Type

import click

from ape import accounts, networks
from ape.api.accounts import AccountAPI
from ape.cli.choices import Alias
from ape.cli.utils import Abort
from ape.exceptions import AliasAlreadyInUseError
from ape.logging import LogLevel, logger


class ApeCliContextObject:
    """A class that can be auto-imported into a plugin ``click.command()``
    via ``@ape_cli_context()``. It can help do common CLI tasks such as log
    messages to the user or abort execution."""

    def __init__(self):
        self.logger = logger

    @staticmethod
    def abort(msg: str, base_error: Exception = None):
        if base_error:
            logger.error(msg)
            raise Abort(msg) from base_error
        raise Abort(msg)


def verbosity_option(cli_logger):
    """A decorator that adds a `--verbosity, -v` option to the decorated
    command.
    """

    level_names = [lvl.name for lvl in LogLevel]
    names_str = f"{', '.join(level_names[:-1])}, or {level_names[-1]}"

    def decorator(f):
        def _set_level(ctx, param, value):
            log_level = getattr(LogLevel, value.upper(), None)
            if log_level is None:
                raise click.BadParameter(f"Must be one of {names_str}, not {value}")

            cli_logger.setLevel(log_level.name)

        return click.option(
            "--verbosity",
            "-v",
            callback=_set_level,
            default=LogLevel.INFO.name,
            metavar="LVL",
            expose_value=False,
            help=f"One of {names_str}",
            is_eager=True,
        )(f)

    return decorator


def ape_cli_context():
    def decorator(f):
        f = verbosity_option(logger)(f)
        f = click.make_pass_decorator(ApeCliContextObject, ensure=True)(f)
        return f

    return decorator


class NetworkChoice(click.Choice):
    """Wraps ``click.Choice`` to provide network choice defaults for the active project."""

    def __init__(self, case_sensitive=True):
        super().__init__(list(networks.network_choices), case_sensitive)

    def get_metavar(self, param):
        return "[ecosystem-name][:[network-name][:[provider-name]]]"


network_option = click.option(
    "--network",
    type=NetworkChoice(case_sensitive=False),
    default=networks.default_ecosystem.name,
    help="Override the default network and provider. (see ``ape networks list`` for options)",
    show_default=True,
    show_choices=False,
)


def verbose_option(help=""):
    return click.option(
        "-v",
        "--verbose",
        is_flag=True,
        default=False,
        help=help,
    )


def _require_non_existing_alias(arg):
    if arg in accounts.aliases:
        raise AliasAlreadyInUseError(arg)
    return arg


def existing_alias_argument(account_type: Optional[Type[AccountAPI]] = None):
    return click.argument("alias", type=Alias(account_type=account_type))


non_existing_alias_argument = click.argument(
    "alias", callback=lambda ctx, param, arg: _require_non_existing_alias(arg)
)


def skip_confirmation_option(help=""):
    return click.option(
        "-y",
        "--yes",
        "skip_confirmation",
        default=False,
        is_flag=True,
        help=help,
    )



def test_compile(ape_cli, runner, project):
    if not (project.path / "contracts").exists():
        result = runner.invoke(ape_cli, ["compile"])
        assert result.exit_code == 0
        assert "WARNING" in result.output
        assert "No 'contracts/' directory detected" in result.output
        return  # Nothing else to test for this project

    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    # First time it compiles, it compiles fully
    for file in project.path.glob("contracts/**/*"):
        assert file.stem in result.output

    result = runner.invoke(ape_cli, ["compile"])
    assert result.exit_code == 0
    # First time it compiles, it caches
    for file in project.path.glob("contracts/**/*"):
        assert file.stem not in result.output

    if not any(c.deploymentBytecode for c in project.contracts.values()):
        return  # Only interfaces

    result = runner.invoke(ape_cli, ["compile", "--size"])
    assert result.exit_code == 0
    # Still caches but displays bytecode size
    for file in project.path.glob("contracts/**/*"):
        assert file.stem in result.output



from . import networks
from .base import abstractdataclass, abstractmethod


@abstractdataclass
class ExplorerAPI:
    """
    An Explorer must work with a particular Network in a particular Ecosystem
    """

    name: str  # Plugin name
    network: networks.NetworkAPI
    request_header: str

    @abstractmethod
    def get_address_url(self, address: str) -> str:
        ...

    @abstractmethod
    def get_transaction_url(self, transaction_hash: str) -> str:
        ...



class ApeException(Exception):
    """
    An exception raised by ape.
    """


class AccountsError(ApeException):
    """
    Raised when a problem occurs when using accounts.
    """


class AliasAlreadyInUseError(AccountsError):
    """
    Raised when attempting to add an account using an alias
    that already maps to another account.
    """

    def __init__(self, alias: str):
        self.alias = alias
        super().__init__(f"Account with alias '{alias}' already in use")


class ContractError(ApeException):
    """
    Raised when issues occur when interacting with a contract
    (calls or transactions).
    """


class ContractCallError(ContractError):
    """
    Raised when issues occur when making a contract call.
    """

    def __init__(self, message=None):
        message = message or "Number of args does not match"
        super().__init__(message)


class TransactionError(ContractError):
    """
    Raised when issues occur while making contract transactions.
    """


class DecodingError(ContractError):
    """
    Raised when issues occur while decoding data from
    a contract call or transaction.
    """

    def __init__(self):
        super().__init__("Output corrupted")


class ContractDeployError(ApeException):
    """
    Raised when a problem occurs when deploying a contract.
    """


class NetworkError(ApeException):
    """
    Raised when a problem occurs when using blockchain networks.
    """


class NetworkNotFoundError(NetworkError):
    """
    Raised when the network with the given name was not found.
    """

    def __init__(self, network: str):
        self.network = network
        message = f"No network named '{network}'"
        super().__init__(message)


class CompilerError(ApeException):
    """
    Raised when unable to compile.
    """


class ProjectError(ApeException):
    """
    Raised when problems occur in a project.
    """


class ConversionError(ApeException):
    """
    Raised when unable to convert a value.
    """


class ProviderError(ApeException):
    """
    Raised when a problem occurs when using providers.
    """


class ConfigError(ApeException):
    """
    Raised when a problem occurs from the configuration file.
    """


class AddressError(ApeException):
    """
    Raised when a problem occurs regarding an address.
    """



import json
from pathlib import Path
from typing import List, Set

from ape.api.compiler import CompilerAPI
from ape.exceptions import CompilerError
from ape.types import ContractType


class InterfaceCompiler(CompilerAPI):
    @property
    def name(self) -> str:
        return "ethpm"

    def get_versions(self, all_paths: List[Path]) -> Set[str]:
        # NOTE: This bypasses the serialization of this compiler into the package manifest's
        #       ``compilers`` field. You should not do this with a real compiler plugin.
        return set()

    def compile(self, filepaths: List[Path]) -> List[ContractType]:
        contract_types: List[ContractType] = []
        for path in filepaths:
            with path.open() as f:
                data = json.load(f)

            if not isinstance(data, list):
                raise CompilerError("Not a valid ABI interface JSON file")

            else:
                contract_types.append(
                    ContractType(  # type: ignore
                        contractName=path.stem,
                        abi=data,
                        sourceId=str(path),
                    )
                )

        return contract_types



import json
from pathlib import Path
from typing import Iterator, Optional

import click
from eth_account import Account as EthAccount  # type: ignore

from ape.api import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.convert import to_address
from ape.exceptions import AccountsError
from ape.types import AddressType, MessageSignature, SignableMessage, TransactionSignature


class InvalidPasswordError(AccountsError):
    """
    Raised when password to unlock an account is incorrect.
    """

    def __init__(self):
        super().__init__("Invalid password")


class AccountContainer(AccountContainerAPI):
    @property
    def _keyfiles(self) -> Iterator[Path]:
        return self.data_folder.glob("*.json")

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._keyfiles:
            yield p.stem

    def __len__(self) -> int:
        return len([*self._keyfiles])

    def __iter__(self) -> Iterator[AccountAPI]:
        for keyfile in self._keyfiles:
            yield KeyfileAccount(self, keyfile)  # type: ignore


# NOTE: `AccountAPI` is a dataclass
class KeyfileAccount(AccountAPI):
    _keyfile: Path

    def __post_init__(self):
        self.locked = True
        self.__cached_key = None

    @property
    def alias(self) -> str:
        return self._keyfile.stem

    @property
    def keyfile(self) -> dict:
        return json.loads(self._keyfile.read_text())

    @property
    def address(self) -> AddressType:
        return to_address(self.keyfile["address"])

    @property
    def __key(self) -> EthAccount:
        if self.__cached_key is not None:
            if not self.locked:
                click.echo(f"Using cached key for '{self.alias}'")
                return self.__cached_key
            else:
                self.__cached_key = None

        passphrase = click.prompt(
            f"Enter Passphrase to unlock '{self.alias}'",
            hide_input=True,
            default="",  # Just in case there's no passphrase
        )

        try:
            key = EthAccount.decrypt(self.keyfile, passphrase)

        except ValueError as err:
            raise InvalidPasswordError() from err

        if click.confirm(f"Leave '{self.alias}' unlocked?"):
            self.locked = False
            self.__cached_key = key

        return key

    def unlock(self):
        passphrase = click.prompt(
            f"Enter Passphrase to permanently unlock '{self.alias}'",
            hide_input=True,
        )

        try:
            self.__cached_key = EthAccount.decrypt(self.keyfile, passphrase)

        except ValueError as err:
            raise InvalidPasswordError() from err

    def lock(self):
        self.locked = True

    def change_password(self):
        self.locked = True  # force entering passphrase to get key
        key = self.__key

        passphrase = click.prompt(
            "Create New Passphrase",
            hide_input=True,
            confirmation_prompt=True,
        )

        self._keyfile.write_text(json.dumps(EthAccount.encrypt(key, passphrase)))

    def delete(self):
        passphrase = click.prompt(
            f"Enter Passphrase to delete '{self.alias}'",
            hide_input=True,
            default="",  # Just in case there's no passphrase
        )

        EthAccount.decrypt(self.keyfile, passphrase)

        self._keyfile.unlink()

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        if self.locked and not click.confirm(f"{msg}\n\nSign: "):
            return None

        signed_msg = EthAccount.sign_message(msg, self.__key)
        return MessageSignature(v=signed_msg.v, r=signed_msg.r, s=signed_msg.s)  # type: ignore

    def sign_transaction(self, txn: TransactionAPI) -> Optional[TransactionSignature]:
        if self.locked and not click.confirm(f"{txn}\n\nSign: "):
            return None

        signed_txn = EthAccount.sign_transaction(txn.as_dict(), self.__key)
        return TransactionSignature(v=signed_txn.v, r=signed_txn.r, s=signed_txn.s)  # type: ignore



BAD_COMMAND = "not-a-name"


def test_run(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run"])
    assert result.exit_code == 1
    assert "Must provide at least one script name or path" in result.output

    result = runner.invoke(ape_cli, ["run", BAD_COMMAND])
    assert result.exit_code == 1
    if not (project.path / "scripts").exists():
        assert "No `scripts/` directory detected to run script" in result.output

    else:
        assert f"No script named '{BAD_COMMAND}' detected in scripts folder" in result.output

    for script_file in (project.path / "scripts").glob("*.py"):
        result = runner.invoke(
            ape_cli, ["run", script_file.stem, "--network", "ethereum:development:test"]
        )
        assert result.exit_code == 0



from typing import Any, Callable, TypeVar, cast

import pluggy  # type: ignore

F = TypeVar("F", bound=Callable[..., Any])
hookimpl = cast(Callable[[F], F], pluggy.HookimplMarker("ape"))
hookspec = pluggy.HookspecMarker("ape")

plugin_manager = pluggy.PluginManager("ape")


class PluginType:
    pass



from decimal import Decimal

from ape.api import ConverterAPI

ETHER_UNITS = {
    "eth": int(1e18),
    "ether": int(1e18),
    "milliether": int(1e15),
    "finney": int(1e15),
    "microether": int(1e12),
    "szabo": int(1e12),
    "gwei": int(1e9),
    "shannon": int(1e9),
    "mwei": int(1e6),
    "lovelace": int(1e6),
    "kwei": int(1e3),
    "babbage": int(1e3),
    "wei": 1,
}


class WeiConversions(ConverterAPI):
    """Converts units like `1 ether` to 1e18 wei"""

    def is_convertible(self, value: str) -> bool:
        if not isinstance(value, str):
            return False

        if " " not in value or len(value.split(" ")) > 2:
            return False

        _, unit = value.split(" ")

        return unit.lower() in ETHER_UNITS

    def convert(self, value: str) -> int:
        value, unit = value.split(" ")

        return int(Decimal(value) * ETHER_UNITS[unit.lower()])



#!/usr/bin/env python
from setuptools import find_packages, setup  # type: ignore

extras_require = {
    "test": [  # `test` GitHub Action jobs uses this
        "pytest>=6.0,<7.0",  # Core testing package
        "pytest-xdist",  # multi-process runner
        "pytest-cov",  # Coverage analyzer plugin
        "pytest-mock",  # For creating mocks
        "hypothesis>=6.2.0,<7.0",  # Strategy-based fuzzer
        "hypothesis-jsonschema==0.19.0",  # JSON Schema fuzzer extension
    ],
    "lint": [
        "black>=20.8b1,<21.0",  # auto-formatter and linter
        "mypy>=0.800,<0.900",  # Static type analyzer
        "flake8>=3.8.3,<4.0",  # Style linter
        "flake8-breakpoint>=1.1.0,<2.0.0",  # detect breakpoints left in code
        "flake8-print>=4.0.0,<5.0.0",  # detect print statements left in code
        "isort>=5.7.0,<6.0",  # Import sorting linter
    ],
    "doc": [
        "Sphinx>=3.4.3,<4",  # Documentation generator
        "sphinx_rtd_theme>=0.1.9,<1",  # Readthedocs.org theme
        "towncrier>=19.2.0, <20",  # Generate release notes
        "sphinx-multiversion>=0.2.4,<0.3.0",  # build multiple versions of docs from tags & branches
        "sphinx-autoapi<=1.9.0",  # for better generation of docs beyond the built-in `autogen`
    ],
    "release": [  # `release` GitHub Action job uses this
        "setuptools",  # Installation tool
        "wheel",  # Packaging tool
        "twine",  # Package upload tool
    ],
    "dev": [
        "commitizen",  # Manage commits and publishing releases
        "pre-commit",  # Ensure that linters are run prior to committing
        "pytest-watch",  # `ptw` test watcher/runner
        "ipdb",  # Debugger (Must use `export PYTHONBREAKPOINT=ipdb.set_trace`)
    ],
}

# NOTE: `pip install -e .[dev]` to install package
extras_require["dev"] = (
    extras_require["test"]
    + extras_require["lint"]
    + extras_require["doc"]
    + extras_require["release"]
    + extras_require["dev"]
)

with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="eth-ape",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Ape Ethereum Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ApeWorX Ltd.",
    author_email="admin@apeworx.io",
    url="https://github.com/ApeWorX/ape",
    include_package_data=True,
    install_requires=[
        "backports.cached_property ; python_version<'3.8'",
        "click>=8.0.0",
        "dataclassy>=0.10.3,<1.0",
        "eth-account>=0.5.5,<0.6.0",
        "pluggy>=0.13.1,<1.0",
        "PyGithub>=1.54,<2.0",
        "pyyaml>=0.2.5",
        "importlib-metadata",
        "singledispatchmethod ; python_version<'3.8'",
        "IPython>=7.25",
        "web3[tester]>=5.18.0,<6.0.0",
    ],
    entry_points={
        "console_scripts": ["ape=ape._cli:cli"],
        "ape_cli_subcommands": [
            "ape_accounts=ape_accounts._cli:cli",
            "ape_compile=ape_compile._cli:cli",
            "ape_console=ape_console._cli:cli",
            "ape_plugins=ape_plugins._cli:cli",
            "ape_run=ape_run._cli:cli",
            "ape_networks=ape_networks._cli:cli",
        ],
    },
    python_requires=">=3.7,<3.10",
    extras_require=extras_require,
    py_modules=[
        "ape",
        "ape_accounts",
        "ape_compile",
        "ape_console",
        "ape_ethereum",
        "ape_http",
        "ape_networks",
        "ape_plugins",
        "ape_run",
        "ape_test",
        "ape_pm",
    ],
    license="Apache-2.0",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={
        "ape": ["py.typed"],
        "ape_accounts": ["py.typed"],
        "ape_compile": ["py.typed"],
        "ape_ethereum": ["py.typed"],
        "ape_http": ["py.typed"],
        "ape_run": ["py.typed"],
        "ape_test": ["py.typed"],
        "ape_pm": ["py.typed"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



from typing import Iterator, Union

from dataclassy import dataclass
from eth_account.messages import SignableMessage  # type: ignore
from eth_utils import to_bytes


@dataclass(frozen=True, slots=True, kwargs=True)
class _Signature:
    v: int
    r: bytes
    s: bytes

    def __iter__(self) -> Iterator[Union[int, bytes]]:
        # NOTE: Allows tuple destructuring
        yield self.v
        yield self.r
        yield self.s

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} v={self.v} r={self.r.hex()} s={self.s.hex()}>"

    def encode_vrs(self) -> bytes:
        return to_bytes(self.v) + self.r + self.s

    def encode_rsv(self) -> bytes:
        return self.r + self.s + to_bytes(self.v)


class MessageSignature(_Signature):
    pass


class TransactionSignature(_Signature):
    pass


_ = SignableMessage



from .accounts import AccountAPI, AccountContainerAPI
from .address import Address, AddressAPI
from .contracts import ContractLog
from .convert import ConverterAPI
from .explorers import ExplorerAPI
from .networks import EcosystemAPI, NetworkAPI, ProviderContextManager, create_network_type
from .providers import ProviderAPI, ReceiptAPI, TransactionAPI, TransactionStatusEnum

__all__ = [
    "AccountAPI",
    "AccountContainerAPI",
    "Address",
    "AddressAPI",
    "ContractInstance",
    "ContractLog",
    "ConverterAPI",
    "EcosystemAPI",
    "ExplorerAPI",
    "ProviderAPI",
    "ProviderContextManager",
    "NetworkAPI",
    "ReceiptAPI",
    "TransactionAPI",
    "TransactionStatusEnum",
    "create_network_type",
]



import json
from pathlib import Path
from typing import Dict, List, Optional

import requests
from dataclassy import dataclass

from ape.types import Checksum, Compiler, ContractType, PackageManifest, Source
from ape.utils import compute_checksum

from ..exceptions import ProjectError
from .compilers import CompilerManager
from .config import ConfigManager


@dataclass
class ProjectManager:
    path: Path
    config: ConfigManager
    compilers: CompilerManager

    dependencies: Dict[str, PackageManifest] = dict()

    def __post_init__(self):
        if isinstance(self.path, str):
            self.path = Path(self.path)

        self.dependencies = {
            manifest.name: manifest
            for manifest in map(self._extract_manifest, self.config.dependencies)
        }

    def _extract_manifest(self, manifest_uri: str) -> PackageManifest:
        manifest_dict = requests.get(manifest_uri).json()
        # TODO: Handle non-manifest URLs e.g. Ape/Brownie projects, Hardhat/Truffle projects, etc.
        if "name" not in manifest_dict:
            raise ProjectError("Dependencies must have a name")

        return PackageManifest.from_dict(manifest_dict)

    def __str__(self) -> str:
        return f'Project("{self.path}")'

    @property
    def _cache_folder(self) -> Path:
        folder = self.path / ".build"
        # NOTE: If we use the cache folder, we expect it to exist
        folder.mkdir(exist_ok=True)
        return folder

    @property
    def manifest_cachefile(self) -> Path:
        file_name = self.config.name or "__local__"
        return self._cache_folder / (file_name + ".json")

    @property
    def cached_manifest(self) -> Optional[PackageManifest]:
        manifest_file = self.manifest_cachefile
        if manifest_file.exists():
            manifest_json = json.loads(manifest_file.read_text())
            if "manifest" not in manifest_json:
                raise ProjectError("Corrupted manifest")

            return PackageManifest.from_dict(manifest_json)

        else:
            return None

    # NOTE: Using these paths should handle the case when the folder doesn't exist
    @property
    def _contracts_folder(self) -> Path:
        return self.path / "contracts"

    @property
    def sources(self) -> List[Path]:
        """All the source files in the project.
        Excludes files with extensions that don't have a registered compiler.
        """
        files: List[Path] = []
        for extension in self.compilers.registered_compilers:
            files.extend(self._contracts_folder.rglob("*" + extension))

        return files

    def load_contracts(self, use_cache: bool = True) -> Dict[str, ContractType]:
        # Load a cached or clean manifest (to use for caching)
        manifest = use_cache and self.cached_manifest or PackageManifest()
        cached_sources = manifest.sources or {}
        cached_contract_types = manifest.contractTypes or {}

        # If a file is deleted from ``self.sources`` but is in
        # ``cached_sources``, remove its corresponding ``contract_types`` by
        # using ``ContractType.sourceId`` and ``ContractType.sourcePath``
        deleted_sources = cached_sources.keys() - set(map(str, self.sources))
        contract_types = {}
        for name, ct in cached_contract_types.items():
            if ct.sourceId in deleted_sources:
                pass  # this contract's source code file was deleted
            else:
                contract_types[name] = ct

        def file_needs_compiling(source: Path) -> bool:
            path = str(source)
            # New file added?
            if path not in cached_sources:
                return True

            # Recalculate checksum if it doesn't exist yet
            cached = cached_sources[path]
            cached.compute_checksum(algorithm="md5")

            assert cached.checksum  # to tell mypy this can't be None

            # File contents changed in source code folder?
            checksum = compute_checksum(
                source.read_bytes(),
                algorithm=cached.checksum.algorithm,
            )
            return checksum != cached.checksum.hash

        # NOTE: filter by checksum, etc., and compile what's needed
        #       to bring our cached manifest up-to-date
        needs_compiling = filter(file_needs_compiling, self.sources)
        contract_types.update(self.compilers.compile(list(needs_compiling)))

        # Update cached contract types & source code entries in cached manifest
        manifest.contractTypes = contract_types
        cached_sources = {
            str(source): Source(  # type: ignore
                checksum=Checksum(  # type: ignore
                    algorithm="md5", hash=compute_checksum(source.read_bytes())
                ),
                urls=[],
            )
            for source in self.sources
        }
        manifest.sources = cached_sources

        # NOTE: Cache the updated manifest to disk (so ``self.cached_manifest`` reads next time)
        self.manifest_cachefile.write_text(json.dumps(manifest.to_dict()))

        return contract_types

    @property
    def contracts(self) -> Dict[str, ContractType]:
        return self.load_contracts()

    def __getattr__(self, attr_name: str):
        contracts = self.load_contracts()
        if attr_name in contracts:
            return contracts[attr_name]
        elif attr_name in self.dependencies:
            return self.dependencies[attr_name]
        else:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{attr_name}'")

    @property
    def _interfaces_folder(self) -> Path:
        return self.path / "interfaces"

    @property
    def _scripts_folder(self) -> Path:
        return self.path / "scripts"

    @property
    def _tests_folder(self) -> Path:
        return self.path / "tests"

    # TODO: Make this work for generating and caching the manifest file
    @property
    def compiler_data(self) -> List[Compiler]:
        compilers = []

        for extension, compiler in self.compilers.registered_compilers.items():
            for version in compiler.get_versions(
                [p for p in self.sources if p.suffix == extension]
            ):
                compilers.append(Compiler(compiler.name, version))  # type: ignore

        return compilers

    # @property
    # def meta(self) -> PackageMeta:
    #     return PackageMeta(**self.config.get_config("ethpm").serialize())

    # def publish_manifest(self):
    #     manifest = self.manifest.to_dict()  # noqa: F841
    #     if not manifest["name"]:
    #         raise ProjectError("Need name to release manifest")
    #     if not manifest["version"]:
    #         raise ProjectError("Need version to release manifest")
    #     # TODO: Clean up manifest and minify it
    #     # TODO: Publish sources to IPFS and replace with CIDs
    #     # TODO: Publish to IPFS



from typing import Iterator

from web3 import EthereumTesterProvider, Web3

from ape.api import ProviderAPI, ReceiptAPI, TransactionAPI


class LocalNetwork(ProviderAPI):
    _web3: Web3 = None  # type: ignore

    def connect(self):
        pass

    def disconnect(self):
        pass

    def update_settings(self, new_settings: dict):
        pass

    def __post_init__(self):
        self._web3 = Web3(EthereumTesterProvider())

    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
        return self._web3.eth.estimate_gas(txn.as_dict())  # type: ignore

    @property
    def chain_id(self) -> int:
        return self._web3.eth.chain_id

    @property
    def gas_price(self):
        # NOTE: Test chain doesn't care about gas prices
        return 0

    def get_nonce(self, address: str) -> int:
        return self._web3.eth.get_transaction_count(address)  # type: ignore

    def get_balance(self, address: str) -> int:
        return self._web3.eth.get_balance(address)  # type: ignore

    def get_code(self, address: str) -> bytes:
        return self._web3.eth.get_code(address)  # type: ignore

    def send_call(self, txn: TransactionAPI) -> bytes:
        data = txn.as_dict()
        if data["gas"] == 0:
            data["gas"] = int(1e12)
        return self._web3.eth.call(data)

    def get_transaction(self, txn_hash: str) -> ReceiptAPI:
        # TODO: Work on API that let's you work with ReceiptAPI and re-send transactions
        receipt = self._web3.eth.wait_for_transaction_receipt(txn_hash)  # type: ignore
        txn = self._web3.eth.get_transaction(txn_hash)  # type: ignore
        return self.network.ecosystem.receipt_class.decode({**txn, **receipt})

    def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
        txn_hash = self._web3.eth.send_raw_transaction(txn.encode())
        return self.get_transaction(txn_hash.hex())

    def get_events(self, **filter_params) -> Iterator[dict]:
        return iter(self._web3.eth.get_logs(filter_params))  # type: ignore



import collections
import json
import os
from copy import deepcopy
from functools import lru_cache
from hashlib import md5
from pathlib import Path
from typing import Any, Dict

import yaml
from importlib_metadata import PackageNotFoundError, packages_distributions, version

from ape.logging import logger

try:
    from functools import cached_property  # type: ignore
except ImportError:
    from backports.cached_property import cached_property  # type: ignore

try:
    from functools import singledispatchmethod  # type: ignore
except ImportError:
    from singledispatchmethod import singledispatchmethod  # type: ignore


@lru_cache(maxsize=None)
def get_distributions():
    return packages_distributions()


def is_relative_to(path: Path, target: Path) -> bool:
    if hasattr(path, "is_relative_to"):
        # NOTE: Only available ``>=3.9``
        return target.is_relative_to(path)  # type: ignore

    else:
        try:
            return target.relative_to(path) is not None
        except ValueError:
            return False


def get_relative_path(target: Path, anchor: Path) -> Path:
    """
    Compute the relative path of ``target`` relative to ``anchor``,
    which may or may not share a common ancestor.
    NOTE: Both paths must be absolute
    """
    if not target.is_absolute():
        raise ValueError("'target' must be an absolute path")
    if not anchor.is_absolute():
        raise ValueError("'anchor' must be an absolute path")

    anchor_copy = Path(str(anchor))
    levels_deep = 0
    while not is_relative_to(anchor_copy, target):
        levels_deep += 1
        anchor_copy = anchor_copy.parent

    return Path("/".join(".." for _ in range(levels_deep))).joinpath(
        str(target.relative_to(anchor_copy))
    )


def get_package_version(obj: Any) -> str:
    # If value is already cached/static
    if hasattr(obj, "__version__"):
        return obj.__version__

    # NOTE: In case were don't pass a module name
    if not isinstance(obj, str):
        obj = obj.__name__

    # Reduce module string to base package
    # NOTE: Assumed that string input is module name e.g. ``__name__``
    pkg_name = obj.split(".")[0]

    # NOTE: In case the distribution and package name differ
    dists = get_distributions()
    if pkg_name in dists:
        # NOTE: Shouldn't really be more than 1, but never know
        if len(dists[pkg_name]) != 1:
            logger.warning(f"duplicate pkg_name '{pkg_name}'")
        pkg_name = dists[pkg_name][0]

    try:
        return version(pkg_name)

    except PackageNotFoundError:
        # NOTE: Must handle empty string result here
        return ""


def deep_merge(dict1, dict2):
    """Return a new dictionary by merging two dictionaries recursively."""

    result = deepcopy(dict1)

    for key, value in dict2.items():
        if isinstance(value, collections.Mapping):
            result[key] = deep_merge(result.get(key, {}), value)
        else:
            result[key] = deepcopy(dict2[key])

    return result


def expand_environment_variables(contents: str) -> str:
    return os.path.expandvars(contents)


def load_config(path: Path, expand_envars=True, must_exist=False) -> Dict:
    if path.exists():
        contents = path.read_text()
        if expand_envars:
            contents = expand_environment_variables(contents)

        if path.suffix in (".json",):
            config = json.loads(contents)
        elif path.suffix in (".yml", ".yaml"):
            config = yaml.safe_load(contents)
        else:
            raise TypeError(f"Cannot parse '{path.suffix}' files!")

        return config or {}

    elif must_exist:
        raise OSError(f"{path} does not exist!")

    else:
        return {}


def compute_checksum(source: bytes, algorithm: str = "md5") -> str:
    if algorithm == "md5":
        hasher = md5
    else:
        raise ValueError(f"Unknown algorithm `{algorithm}`")

    return hasher(source).hexdigest()


__all__ = [
    "cached_property",
    "deep_merge",
    "expand_environment_variables",
    "load_config",
    "singledispatchmethod",
]
